"""
Evaluate fitness from parent and offspring results of 1st generation.
Select best topologies by pareto-front condition.
Export these as 2nd generation; "Topologies_2" and "FieldOutput_2".
"""


def run():
    from ..PostProcessing import evaluate_all_fitness_values, selection
    from ..ParameterDefinitions import Parameters, fitness_definitions
    from ..FileIO import pickle_io
    import numpy as np
    from dataclasses import asdict
    import os
    script_dir = os.path.dirname(__file__)
    sample_data_dir = os.path.join(script_dir, 'sample_data')
    os.chdir(sample_data_dir)

    # Define fitness value evaluation version. Definitions are organized in auxeticmop.ParameterDefinitions.fitness_definitions
    evaluation_version = 'ver3'

    # Initialize parameters
    parameters = Parameters(evaluation_version=evaluation_version)
    parameters.post_initialize()

    # Load topologies and results(field outputs from ABAQUS)
    loaded_topologies = pickle_io('Topologies_1', mode='r')
    parent_topologies = loaded_topologies['parent']
    offspring_topologies = loaded_topologies['offspring']
    parent_results = pickle_io('FieldOutput_1', mode='r')
    offspring_results = pickle_io('FieldOutput_offspring_1', mode='r')
    assert len(parent_topologies) == len(offspring_topologies) == len(parent_results) == len(offspring_results)

    # Evaluate fitness values
    parent_fitness_values = evaluate_all_fitness_values(fitness_definitions=fitness_definitions,
                                                        params_dict=asdict(parameters),
                                                        results=parent_results, topologies=parent_topologies)
    offspring_fitness_values = evaluate_all_fitness_values(fitness_definitions=fitness_definitions,
                                                           params_dict=asdict(parameters),
                                                           results=offspring_results, topologies=offspring_topologies)
    print('Shape of parent fitness values: ', parent_fitness_values.shape)
    print('Shape of offspring fitness values: ', offspring_fitness_values.shape)

    # Gather topologies, results, and fitness values
    all_topologies = np.vstack((parent_topologies, offspring_topologies))
    all_results = parent_results.copy()
    all_results.update({entity_num + len(parent_results): offspring_results[entity_num]
                        for entity_num in sorted(offspring_results.keys())})
    all_fitness_values = evaluate_all_fitness_values(fitness_definitions=fitness_definitions, params_dict=asdict(parameters),
                                                     results=all_results, topologies=all_topologies)
    print('Shape of all topologies: ', all_topologies.shape)
    print('size of all results: ', len(all_results))
    print('Shape of all fitness values: ', all_fitness_values.shape)

    # Topologies of parent of next generation will be selected by pareto fronts criterion
    pareto_indices = selection(all_fitness_values=all_fitness_values, selected_size=parameters.end_pop)
    print('Entity numbers which are selected by pareto-front-criterion: \n', pareto_indices + 1)
    selected_topologies = all_topologies[pareto_indices]
    selected_results = {entity_num: all_results[pareto_idx + 1]
                        for entity_num, pareto_idx in enumerate(pareto_indices, start=1)}
    print('Shape of selected topologies: ', selected_topologies.shape)
    print('Size of selected results: ', len(selected_results))
    pickle_io('Topologies_2', mode='w', to_dump={'parent': selected_topologies})
    pickle_io('FieldOutput_2', mode='w', to_dump=selected_results)


if __name__ == '__main__':
    run()



import tkinter as tk
from tkinter.filedialog import askdirectory
from tkinter import messagebox
import os
import pickle
from dataclasses import asdict
from typing import Union
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.ticker import MaxNLocator
from .PostProcessing import get_datum_hv, get_hv_from_datum_hv, find_pareto_front_points, evaluate_all_fitness_values
from .FileIO import pickle_io, get_sorted_file_numbers_from_pattern
from .ParameterDefinitions import Parameters, GuiParameters, translate_dictionary, radiobutton_name_dict, \
    fitness_definitions

# Define parameters for gui
gui_parameters = GuiParameters()
PARAMETER_FILE_NAME = gui_parameters.parameter_file_name
LEFT_WIDTH = gui_parameters.left_width
RIGHT_WIDTH = gui_parameters.right_width
BUTTON_WIDTH = gui_parameters.button_width
PADX = gui_parameters.padx
PADY = gui_parameters.pady
HEIGHT = gui_parameters.height
POLLING_RATE = gui_parameters.polling_rate
TITLE = gui_parameters.title
RADIOBUTTON_NAME_DICT = radiobutton_name_dict


class App:  # GUI class
    def __init__(self, conn=None):
        # Root configuration
        self.conn = conn
        self.ready_to_run = False
        self.root = tk.Tk()
        self.root.protocol("WM_DELETE_WINDOW", self.callback_quit)
        self.root.title(TITLE)
        self.root.config(background='#FFFFFF')
        self.root.resizable(False, False)
        self.polling_rate = POLLING_RATE

        # Frames
        self.up_frame = tk.Frame(self.root, width=LEFT_WIDTH + RIGHT_WIDTH + 2 * PADX, height=100)
        self.up_frame.config(background='#FFFFFF')
        self.up_frame.grid(row=0, column=0, padx=PADX, pady=PADY / 2)
        self.up_frame.grid_propagate(False)
        self.down_frame = tk.Frame(self.root, width=RIGHT_WIDTH + LEFT_WIDTH + 2 * PADX, height=HEIGHT)
        self.down_frame.config(background='#FFFFFF')
        self.left_frame = tk.Frame(self.down_frame, width=LEFT_WIDTH, height=HEIGHT, padx=PADX, pady=PADY)
        self.right_frame = tk.Frame(self.down_frame, width=RIGHT_WIDTH, height=HEIGHT, padx=PADX, pady=PADY)

        # Elements
        self.set_path_title = tk.Label(self.up_frame,
                                       text='Choose the folder containing the Abaqus script and pickle files.')
        self.set_path_title.config(background='#FFFFFF')
        self.set_path_display = tk.Listbox(self.up_frame, width=50, height=1)
        self.set_path_btn = tk.Button(self.up_frame, text='Browse folder...', width=BUTTON_WIDTH,
                                      command=self.onclick_set_path_button)
        self.submit_btn = tk.Button(self.right_frame, width=BUTTON_WIDTH, text='Save presets',
                                    command=self.onclick_submit_btn)
        self.exit_btn = tk.Button(self.right_frame, width=BUTTON_WIDTH, text='Exit',
                                  command=self.onclick_exit_btn, background='#FF0000', foreground='#FFFFFF')
        self.set_default_btn = tk.Button(self.right_frame, width=BUTTON_WIDTH, text='Load defaults',
                                         command=self.onclick_set_default_btn)
        self.setPath = tk.StringVar()
        self.string_vars = [tk.StringVar() for _ in Parameters.__dict__]
        self.figure, self.ax = plt.subplots(nrows=1, ncols=2,
                                            figsize=((LEFT_WIDTH - PADX) / 100, (HEIGHT - PADY) / 100), dpi=100)
        self.bar = FigureCanvasTkAgg(self.figure, self.left_frame)
        self.bar.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH)
        self.parameters = Parameters()
        self.parameters_dict = asdict(self.parameters)

        # Show main
        self.set_path_title.pack()
        self.set_path_display.pack()
        self.set_path_btn.pack()
        self.root.mainloop()

    def callback_quit(self):
        self.root.quit()

    def show_canvas(self):
        # Show graphs to be plotted
        self.ax[0].set(title='Pareto Fronts', xlabel='Objective function 1', ylabel='Objective function 2')
        self.ax[1].set(title='Hyper Volume by Generation', xlabel='Generation', ylabel='Hyper volume')
        self.ax[0].grid(True)
        self.ax[1].grid(True)
        self.ax[1].xaxis.set_major_locator(MaxNLocator(integer=True))
        if self.conn is not None:
            self.update_canvas()

    def update_canvas(self):
        if self.conn.poll():  # Checking if any received data available for every 1/polling_rate second
            try:
                _x1, _y1, _x2y2 = self.conn.recv()
                _x2, _y2 = zip(*_x2y2.items())
                print('[GUI] Received: ', _x1, _y1, _x2, _y2)

                # Randomize plotting options to make data more noticeable
                _color_1, _color_2 = np.random.rand(3, ), np.random.rand(3, )
                _plot_options = {
                    'marker': np.random.choice(
                        ('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd', 'P', 'X')),
                    'color': _color_1,
                    'markeredgecolor': _color_1,
                    'markerfacecolor': _color_2,
                    'markersize': 8,
                    'markeredgewidth': 2}
                # Delete all plotted hyper volumes
                self.ax[1].clear()
                self.ax[1].set(title='Hyper Volume by Generation', xlabel='Generation', ylabel='Hyper volume')
                self.ax[1].grid(True)
                self.ax[1].xaxis.set_major_locator(MaxNLocator(integer=True))

                # Plot and scatter received data
                self.ax[0].plot(_x1, _y1, **_plot_options)
                for _generation, _hv, _line in zip(_x2, _y2, self.ax[0].lines):
                    self.ax[1].scatter(_generation, _hv, marker=_line.get_marker(), c=[_line.get_markerfacecolor()],
                                       edgecolors=_line.get_markeredgecolor(), s=_line.get_markersize() ** 2,
                                       linewidth=_line.get_markeredgewidth())
                self.bar.draw()
            except Exception as error_message:
                print('[GUI] An plotting error occurred:', error_message)
        self.root.after(int(1000 / self.polling_rate), self.update_canvas)

    def onclick_set_path_button(self):
        try:
            _folder_path = askdirectory(initialdir="./")
            if _folder_path:
                self.set_path_display.delete(0, "end")
                self.set_path_display.insert(0, _folder_path)
            self.setPath.set(self.set_path_display.get(0))
            os.chdir(self.setPath.get())

            _params_already_exists = True if os.path.isfile(PARAMETER_FILE_NAME) else False
            if _params_already_exists:
                with open(PARAMETER_FILE_NAME, mode='rb') as f_params:
                    self.set_path_display.config(background='#00FF00')
                    self.set_path_title.config(text='Preset parameter value found.')
                    self.show_parameters(loaded=pickle.load(f_params))
                    self.ready_to_run = False
            else:
                self.set_path_display.config(background='#FF0000')
                self.set_path_title.config(text='Parameter not found. Please set below.')
                self.show_parameters(loaded=None)
                self.ready_to_run = False

        except Exception as error_message:
            messagebox.showerror("Error", f"An error occurred:\n{error_message}")

    def onclick_set_default_btn(self):
        for _idx, _value in enumerate(self.parameters_dict.values()):
            self.string_vars[_idx].set(_value)

    def return_radiobutton_frame_instead_of_entry(self, key: str, str_var_idx: int, name_dict: dict):  # right_Frame
        _radiobutton_frame = tk.Frame(self.right_frame, width=LEFT_WIDTH - 2 * PADX,
                                      height=HEIGHT / len(self.parameters_dict) - PADY)
        if len(name_dict[key]) == 2:
            _label_width = 25
            _radio_button_width = 9
        elif len(name_dict[key]) == 3:
            _label_width = 24
            _radio_button_width = 5
        else:
            raise ValueError('Please make sure the number of checkboxes is less than 3.')
        _label = tk.Label(_radiobutton_frame, width=_label_width, text=translator(dictionary=translate_dictionary,
                                                                                  s=key, flip=False), anchor='w')
        _label.grid(row=0, column=0)

        for menu_idx, menu in enumerate(name_dict[key]):
            _radio_button = tk.Radiobutton(_radiobutton_frame, text=menu, variable=self.string_vars[str_var_idx],
                                           width=_radio_button_width, value=menu, tristatevalue=menu)
            _radio_button.grid(row=0, column=menu_idx + 1)
        return _radiobutton_frame

    def onclick_submit_btn(self):
        for params_idx, key in enumerate(self.parameters_dict.keys()):
            self.parameters_dict[key] = atoi(self.string_vars[params_idx].get())
        if self.ready_to_run and self.conn is not None:
            self.conn.send((self.setPath.get(), Parameters(**self.parameters_dict)))
            self.submit_btn.config(background='#0000FF', foreground='#FFFFFF', text='Running...')
        else:
            with open(PARAMETER_FILE_NAME, mode='wb') as f_params:
                pickle.dump(self.parameters_dict, f_params)
                print(f'[GUI] Dumping to "{os.getcwd()}" Complete')
                for key, value in self.parameters_dict.items():
                    print(f'- {key}: {value}')
            self.submit_btn.config(background='#00FF00', text='Run')
            self.ready_to_run = True

    def onclick_exit_btn(self):
        print('[GUI] Closing GUI...')
        self.root.quit()

    def show_parameters(self, loaded: Union[None, dict]) -> None:
        self.down_frame.grid(row=1, column=0, padx=PADX, pady=PADY / 2)
        self.down_frame.grid_propagate(False)
        self.left_frame.grid(row=1, column=0, padx=PADX, pady=PADY / 2)
        self.left_frame.grid_propagate(False)
        self.right_frame.grid(row=1, column=1, padx=PADX, pady=PADY / 2)
        self.right_frame.grid_propagate(False)
        for row_idx, key in enumerate(asdict(Parameters()).keys()):
            if key in RADIOBUTTON_NAME_DICT.keys():
                _radio_button_frame = self.return_radiobutton_frame_instead_of_entry(key=key, str_var_idx=row_idx,
                                                                                     name_dict=RADIOBUTTON_NAME_DICT)
                _radio_button_frame.grid(row=row_idx, column=0)

            else:
                _entry_frame = tk.Frame(self.right_frame, width=LEFT_WIDTH - 2 * PADX,
                                        height=HEIGHT / len(asdict(Parameters()).keys()) - PADY)
                _label = tk.Label(_entry_frame, width=25, text=translator(dictionary=translate_dictionary,
                                                                          s=key, flip=False), anchor='w')
                _entry = tk.Entry(_entry_frame, width=25, textvariable=self.string_vars[row_idx])
                _entry_frame.grid(row=row_idx, column=0)
                _label.grid(row=0, column=0)
                _entry.grid(row=0, column=1)
            self.string_vars[row_idx].set('')

        empty_label = tk.Label(self.right_frame)
        empty_label.grid(row=len(asdict(Parameters()).keys()))
        self.submit_btn.grid(row=len(asdict(Parameters()).keys()) + 1)
        self.exit_btn.grid(row=len(asdict(Parameters()).keys()) + 2)
        self.set_default_btn.grid(row=len(asdict(Parameters()).keys()) + 3)

        if loaded:
            for params_idx, (key, value) in enumerate(loaded.items()):
                self.string_vars[params_idx].set(value)
        self.show_canvas()


class Visualizer:
    def __init__(self, conn_to_gui=None):
        self.conn_to_gui = conn_to_gui
        self.ref_x = None
        self.ref_y = None
        self.all_lower_bounds = dict()
        self.all_datum_hv = dict()
        if conn_to_gui is None:
            self.figure, self.axes = plt.subplots(nrows=1, ncols=2,
                                                  figsize=((1400 - 5) / 100, (700 - 5) / 100), dpi=100)
            self.axes[0].set(title='Pareto Fronts', xlabel='Objective function 1', ylabel='Objective function 2')
            self.axes[1].set(title='Hyper Volume by Generation', xlabel='Generation', ylabel='Hyper volume')
            self.axes[1].xaxis.set_major_locator(MaxNLocator(integer=True))

    def plot(self, gen_num: int, pareto_1_sorted: np.ndarray, pareto_2_sorted: np.ndarray,
             use_manual_rp: bool = False, ref_x: float = 0.0, ref_y: float = 0.0) -> None:
        # Determining reference point
        if use_manual_rp:
            self.ref_x, self.ref_y = ref_x, ref_y
        elif (self.ref_x is None) or (self.ref_y is None):
            self.ref_x = pareto_1_sorted[-1]
            self.ref_y = pareto_2_sorted[0]
        else:
            if pareto_1_sorted[-1] > self.ref_x:
                self.ref_x = pareto_1_sorted[-1]
            if pareto_2_sorted[0] > self.ref_y:
                self.ref_y = pareto_2_sorted[0]

        # Calculating hyper volume
        _datum_hv = get_datum_hv(pareto_1_sorted, pareto_2_sorted)
        _lower_bounds = [pareto_1_sorted[0], pareto_2_sorted[-1]]
        self.all_datum_hv.update({gen_num: _datum_hv})
        self.all_lower_bounds.update({gen_num: _lower_bounds})
        _all_hv = {key: get_hv_from_datum_hv(self.all_datum_hv[key], self.all_lower_bounds[key],
                                             ref_x=self.ref_x, ref_y=self.ref_y) for key in self.all_datum_hv.keys()}
        _generations, _hvs = zip(*_all_hv.items())

        # Saving plotting data
        _file_name = '_plotting_data_'
        if os.path.isfile(_file_name):
            with open(_file_name, mode='rb') as f_read:
                _read_data = pickle.load(f_read)
            with open(_file_name, mode='wb') as f_write:
                _read_data.update({gen_num: (pareto_1_sorted, pareto_2_sorted)})
                pickle.dump(_read_data, f_write)
        else:
            with open(_file_name, mode='wb') as f_write:
                pickle.dump({gen_num: (pareto_1_sorted, pareto_2_sorted)}, f_write)

        # Plot data onto GUI
        if self.conn_to_gui is not None:
            self.conn_to_gui.send((pareto_1_sorted, pareto_2_sorted, _all_hv))
        else:
            color_1 = np.random.rand(3, )
            color_2 = np.random.rand(3, )
            plot_options = {
                'marker': np.random.choice(('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd', 'P', 'X')),
                'color': color_1,
                'markeredgecolor': color_1,
                'markerfacecolor': color_2,
                'markersize': 8,
                'markeredgewidth': 2}
            print(f'[VISUALIZE] Generation {gen_num}:')
            print(f'> Objective function 1:{pareto_1_sorted}')
            print(f'> Objective function 2:{pareto_2_sorted}\n')
            self.axes[0].plot(pareto_1_sorted, pareto_2_sorted, **plot_options)
            self.axes[1].clear()
            for generation, hv, line in zip(_generations, _hvs, self.axes[0].lines):
                self.axes[1].scatter(generation, hv, marker=line.get_marker(), c=[line.get_markerfacecolor()],
                                     edgecolors=line.get_markeredgecolor(), s=line.get_markersize() ** 2,
                                     linewidth=line.get_markeredgewidth())
            self.axes[0].grid(True)
            self.axes[1].grid(True)

    def visualize(self, params, gen, use_manual_rp, ref_x=0.0, ref_y=0.0):
        next_parent_topologies = pickle_io(f'Topologies_{gen + 1}', mode='r')['parent']
        next_parent_results = pickle_io(f'FieldOutput_{gen + 1}', mode='r')
        next_parent_fitness_values = evaluate_all_fitness_values(fitness_definitions=fitness_definitions,
                                                                 params_dict=asdict(params),
                                                                 topologies=next_parent_topologies,
                                                                 results=next_parent_results)
        next_parent_pareto_indices = find_pareto_front_points(costs=next_parent_fitness_values, return_index=False)
        self.plot(gen_num=gen,
                  pareto_1_sorted=next_parent_pareto_indices[:, 0], pareto_2_sorted=next_parent_pareto_indices[:, 1],
                  use_manual_rp=use_manual_rp, ref_x=ref_x, ref_y=ref_y)


def atoi(s: str) -> Union[str, int, float]:
    try:
        float(s)
        # s is integer or float type
        try:
            int(s)
            # s is int type
            return int(s)
        except ValueError:
            # s is float type
            return float(s)
    except ValueError:
        # s is not number, just string
        return s


def translator(dictionary: dict, s: str, flip: bool = False) -> str:
    return {value: key for key, value in dictionary.items()}.get(s) if flip else dictionary.get(s)


def plot_previously_plotted_data(visualizer: Visualizer, params: Parameters):
    file_numbers = get_sorted_file_numbers_from_pattern(r'FieldOutput_\d+')
    for file_number in file_numbers:
        visualizer.visualize(gen=file_number-1, params=params, use_manual_rp=False)



import socket
import threading
import json
import os
import pickle
import struct
import multiprocessing as mp
from multiprocessing import connection
from datetime import datetime
from time import sleep
from sys import version_info
from typing import Tuple
try:
    from Queue import Queue
except ImportError:
    from queue import Queue


class Server:
    def __init__(self, host, port, option, run_nonblocking):
        self.host = host
        self.port = port
        self.option = option
        self.q = Queue()
        self.connected_clients = list()
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind((self.host, self.port))
        self.server_socket.listen(5)
        self._default_packet_size = 1024
        self._header_format = '>I'
        self._header_bytes = 4
        if run_nonblocking:
            self._server_th = threading.Thread(target=self.run, daemon=True)
            self._server_th.start()

    def run(self):
        while True:
            try:
                print('[{}] Waiting for a client...'.format(datetime.now()))
                client_socket, client_addr = self.server_socket.accept()
                if version_info.major >= 3:
                    new_th = threading.Thread(target=self._thread_recv, args=(client_socket, client_addr, self.option),
                                              daemon=True)
                else:
                    new_th = threading.Thread(target=self._thread_recv, args=(client_socket, client_addr, self.option))
                    new_th.setDaemon(True)
                new_th.start()
                self.connected_clients.append(client_socket)
            except Exception as e:
                print('<!> Server error:', e)
                break
        self.server_socket.close()

    def send(self, client_socket: socket.socket, data: any) -> bool:
        if self.option == 'pickle':
            serialized_data = pickle.dumps(data, protocol=2)
        else:
            serialized_data = json.dumps(data).encode()
        try:
            print('Sending packets: ', serialized_data)
            client_socket.sendall(struct.pack(self._header_format, len(serialized_data)))
            client_socket.sendall(serialized_data)
            print('[{}] A data sent'.format(datetime.now()))
            return True
        except ConnectionError as e:
            print(e)
            return False

    def recv(self):
        return self.q.get()

    def close(self):
        self.server_socket.close()

    def _thread_recv(self, client_socket, client_addr, option):
        print('[{}] {}:{} has joined!'.format(datetime.now(), client_addr[0], client_addr[1]))
        while True:
            try:  # Trying to receive a data and decode it
                data_size = struct.unpack(self._header_format, client_socket.recv(self._header_bytes))[0]
                remaining_payload_size = data_size
                packets = b''
                while remaining_payload_size != 0:
                    packets += client_socket.recv(remaining_payload_size)
                    remaining_payload_size = data_size - len(packets)
                try:  # Trying to decode received data
                    if option == 'json':
                        received_data = json.loads(packets.decode())
                    else:
                        if version_info.major >= 3:
                            received_data = pickle.loads(packets, encoding='bytes')
                        else:
                            received_data = pickle.loads(packets)
                    print('[{}] Received data: {}'.format(datetime.now(), received_data))
                    self.q.put(received_data)
                except Exception as e2:  # Decoding is failed
                    print('[{}] Loading received data failure: {}'.format(datetime.now(), e2))
                    continue
            except Exception as e1:  # Connection is lost
                print('[{}] Error: {}'.format(datetime.now(), e1))
                break
        self.is_alive = False


class Client:
    def __init__(self, host, port, option, connect):
        self.host = host
        self.port = port
        self.option = option
        self.q = Queue()
        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.is_alive = True
        self._default_packet_size = 1024
        self._header_format = '>I'
        self._header_bytes = 4
        if connect:
            self.connect()

    def connect(self):
        if version_info.major >= 3:
            new_th = threading.Thread(target=self._thread_recv, args=(self.client_socket, self.option), daemon=True)
        else:
            new_th = threading.Thread(target=self._thread_recv, args=(self.client_socket, self.option))
            new_th.setDaemon(True)
        self.client_socket.connect((self.host, self.port))
        print('[{}] Connected to {}:{}'.format(datetime.now(), self.host, self.port))
        new_th.start()

    def send(self, data):
        if self.option == 'pickle':
            serialized_data = pickle.dumps(data, protocol=2)
        else:
            serialized_data = json.dumps(data).encode()
        while True:
            try:
                self.client_socket.sendall(struct.pack(self._header_format, len(serialized_data)))
                self.client_socket.sendall(serialized_data)
                print('[{}] A data sent'.format(datetime.now()))
                break
            except Exception as send_error:
                print('Sending data failed, trying to reconnect to server: ', send_error)
                self.connect()

    def recv(self):
        return self.q.get()

    def close(self):
        self.client_socket.close()

    def _thread_recv(self, client_socket, option):
        while True:
            try:  # Trying to receive a data and decode it
                data_size = struct.unpack(self._header_format, client_socket.recv(self._header_bytes))[0]
                remaining_payload_size = data_size
                packets = b''
                while remaining_payload_size != 0:
                    packets += client_socket.recv(remaining_payload_size)
                    remaining_payload_size = data_size - len(packets)
                try:  # Trying to decode received data
                    if option == 'json':
                        received_data = json.loads(packets.decode())
                    else:
                        if version_info.major >= 3:
                            received_data = pickle.loads(packets, encoding='bytes')
                        else:
                            received_data = pickle.loads(packets)
                    print('[{}] Received data: {}'.format(datetime.now(), received_data))
                    self.q.put(received_data)
                except Exception as e2:  # Decoding is failed
                    print('[{}] Loading received data failure: {}'.format(datetime.now(), e2))
                    continue
            except Exception as e1:  # Connection is lost
                print('[{}] Error: {}'.format(datetime.now(), e1))
                break
        self.is_alive = False
        print('<!> Connection dead')


def make_and_start_process(target: any, duplex: bool = True,
                           daemon: bool = True) -> Tuple[mp.Process, connection.Connection, connection.Connection]:
    """
    Make GUI process and return a process and two Pipe connections between main process and GUI process.
    :param target: The GUI class to run as another process.
    :param duplex: If True, both receiving and sending data between main process and GUI process will be allowed.
    Otherwise, conn_1 is only allowed for receiving data and conn_2 is only allowed for sending data.
    :param daemon: If True, GUI process will be terminated when main process is terminated.
    Otherwise, GUI process will be orphan process.
    :return: A running process, Pipe connections of main process, GUI process, respectively.
    """
    conn_1, conn_2 = mp.Pipe(duplex=duplex)
    process = mp.Process(target=target, args=(conn_2,), daemon=daemon)
    process.start()
    return process, conn_1, conn_2


def start_abaqus_cae(option: str) -> mp.Process:
    """
    Open an abaqus CAE process
    :param option: 'noGUI' for abaqus non-gui mode, 'script' for abaqus gui mode.
    :return: Abaqus process.
    """
    print(f"========== Opening ABAQUS on {datetime.now().strftime('%Y/%m/%d %H:%M:%S')}! ==========")
    script_path = os.path.join(os.path.dirname(__file__), 'AbaqusScripts.py')
    process = mp.Process(target=os.system, args=(f'abaqus cae {option}={script_path}',), daemon=True)
    process.start()  # Start abaqus
    return process


def request_abaqus(dict_data: dict, server: Server) -> None:
    """
    Send Json data to ABAQUS
    :param dict_data: Dictionary data to send to ABAQUS
    :param server: A server to ABAQUS
    :return: Nothing
    """
    while len(server.connected_clients) == 0:
        print('Waiting for ABAQUS socket connection ...')
        sleep(1.0)
    is_data_sent = False
    while not is_data_sent:
        is_data_sent = server.send(client_socket=server.connected_clients[-1], data=dict_data)
    print('Waiting for message from ABAQUS ...')
    message_from_client = server.recv()
    print(message_from_client)
    print(f"========== An abaqus job done on {datetime.now().strftime('%Y/%m/%d %H:%M:%S')}! ==========")


# # This function is for Python2. Checkout ABQ.py
# def ascii_encode_dict(data):
#     ascii_encode = lambda x: x.encode('ascii') if isinstance(x, unicode) else x
#     return dict(map(ascii_encode, pair) for pair in data.items())


if __name__ == '__main__':
    open_server = False
    if open_server:  # Creating server
        my_server = Server(host='', port=12345, option='json', run_nonblocking=True)
        print('server created')
        while len(my_server.connected_clients) == 0:
            sleep(1)
        while True:
            data_to_send = {'a': 1, 'b': 2.0, 'c': 'hello', 'd': True, 'e': [0, 1, 2, 3]}
            json_data = json.dumps(data_to_send)
            my_server.send(client_socket=my_server.connected_clients[-1], data=data_to_send)
            print('sending: ', json_data)
            sleep(5)

    else:  # Creating client
        my_client = Client(host='localhost', port=12345, option='json', connect=True)
        while True:
            print('Received: ', my_client.recv())



"""
Generate offspring from parent topologies
"""


def run():
    from ..GeneticAlgorithm import generate_offspring
    from ..ParameterDefinitions import Parameters
    from ..PostProcessing import visualize_one_cube
    from ..FileIO import pickle_io
    import os
    script_dir = os.path.dirname(__file__)
    sample_data_dir = os.path.join(script_dir, 'sample_data')
    os.chdir(sample_data_dir)

    # Parameters
    number_of_voxels_x = 5
    number_of_voxels_y = 5
    number_of_voxels_z = 5
    number_of_topologies_to_create = 10
    mutation_rate = 0.1
    parent_topology_file_name = 'Topologies_1'
    save_offspring_topologies_file_as = 'Topologies_1'
    show_created_offspring_topologies = True

    # Initialize parameters
    parameters = Parameters(lx=number_of_voxels_x, ly=number_of_voxels_y, lz=number_of_voxels_z,
                            end_pop=number_of_topologies_to_create, mutation_rate=mutation_rate)
    parameters.post_initialize()

    # Load parent topologies from pickle file
    parent_topologies = pickle_io(parent_topology_file_name, mode='r')['parent']

    # Create offspring topologies and visualize as full-sized cube
    offspring_topologies = generate_offspring(gen=1, topo_parents=parent_topologies, params=parameters,
                                              save_file_as=save_offspring_topologies_file_as)
    if show_created_offspring_topologies:
        for offspring_topology in offspring_topologies:
            visualize_one_cube(cube_3d_array=offspring_topology, full=True)


if __name__ == '__main__':
    run()



import os
import pickle
import asyncio
import aiofiles
import numpy as np
import re
from typing import Union


def remove_file(file_name: str) -> None:
    """
    Remove a file with designated filename.
    :param file_name: The filename to remove.
    :return: None
    """
    if os.path.isfile(file_name):
        os.remove(file_name)


def key_modifier(key: str, option: Union[str, None]) -> object:
    if option == 'int':
        return int(re.compile(r'\d+').search(key).group())
    else:
        return key


def get_sorted_file_numbers_from_pattern(p):
    return sorted(
        [int(re.compile(r'\d+').search(s).group()) for s in [f for f in os.listdir() if re.compile(p).match(f)]])


def find_job_location_from_offspring(params_dict):
    from .PostProcessing import evaluate_all_fitness_values, find_pareto_front_points
    from .ParameterDefinitions import fitness_definitions

    def _find_job_location_from_offspring(g, prt_idx, tp_offs, tp_prt, tp_pf):
        arg = np.argwhere(np.all(tp_offs == tp_prt, axis=(2,)))
        if len(arg) == 0:
            arg = np.argwhere(np.all(tp_pf == tp_prt, axis=(1,)))
            print(f'Pareto topo in parent {g} - {prt_idx + 1} is in parent ', 1, '-', arg[0, 0] + 1)
        else:
            print(f'Pareto topo in parent {g} - {prt_idx + 1} is in offspring ', arg[0, 0] + 1, '-',
                  arg[0, 1] + 1)

    file_headers = ('Topologies', 'FieldOutput_offspring', 'FieldOutput')
    topo_parent, topo_offspring, result_parent, result_offspring = None, None, dict(), dict()
    for file_header in file_headers:
        file_numbers = get_sorted_file_numbers_from_pattern(rf'{file_header}_\d+')
        file_names = [f'{file_header}_{file_number}' for file_number in file_numbers]
        loaded_dict = asyncio.run(pickles_aio(file_names, mode='r', key_option='int'))
        for gen_num in loaded_dict.keys():
            if 'parent' in loaded_dict[gen_num]:
                if topo_parent is None:
                    topo_parent = np.expand_dims(loaded_dict[gen_num]['parent'], axis=0)
                else:
                    topo_parent = np.vstack((topo_parent, np.expand_dims(loaded_dict[gen_num]['parent'], axis=0)))
            if 'offspring' in loaded_dict[gen_num]:
                if topo_offspring is None:
                    topo_offspring = np.expand_dims(loaded_dict[gen_num]['offspring'], axis=0)
                else:
                    topo_offspring = np.vstack(
                        (topo_offspring, np.expand_dims(loaded_dict[gen_num]['offspring'], axis=0)))
            else:
                if file_header == 'FieldOutput':
                    result_parent.update({gen_num: loaded_dict[gen_num]})
                elif file_header == 'FieldOutput_offspring':
                    result_offspring.update({gen_num: loaded_dict[gen_num]})
    for gen in get_sorted_file_numbers_from_pattern(r'FieldOutput_\d+'):
        fit_vals = evaluate_all_fitness_values(params_dict=params_dict, results=result_parent[gen],
                                               topologies=topo_parent[gen - 1], fitness_definitions=fitness_definitions)
        pareto_indices = find_pareto_front_points(costs=fit_vals, return_index=True)
        print(f'\n>>> Parent {gen} <<<')
        print('Pareto topos: ', pareto_indices + 1)
        for pareto_index in pareto_indices:
            topo_pareto = topo_parent[gen - 1][pareto_index]
            _find_job_location_from_offspring(g=gen, prt_idx=pareto_index, tp_offs=topo_offspring,
                                              tp_prt=topo_pareto, tp_pf=topo_parent[0])


# def dump_pickled_dict_data(file_name: str, key: object, to_dump: object, mode: str) -> None:
#     if mode == 'a' and os.path.isfile(file_name):
#         with open(file_name, mode='rb') as f:
#             dict_data = pickle.load(f, encoding='latin1')
#         dict_data.update({key: to_dump})
#     else:
#         dict_data = {key: to_dump}
#     with open(file_name, mode='wb') as f:
#         pickle.dump(dict_data, f, protocol=2)
#
#
# def load_pickled_dict_data(file_name: str) -> dict:
#     with open(file_name, mode='rb') as f:
#         dict_data = pickle.load(f, encoding='latin1')
#     return dict_data


async def pickle_aio(file_name: str, mode: str, to_dump: Union[dict, list, np.ndarray] = None) -> any:
    encoding = 'latin1'
    if mode == 'r':
        async with aiofiles.open(file_name, mode='rb') as f:
            serialized_pickle = await f.read()
        print(file_name, 'loaded!')
        return pickle.loads(serialized_pickle, encoding=encoding)
    elif mode == 'a' and os.path.isfile(file_name):
        async with aiofiles.open(file_name, mode='rb') as f:
            serialized_pickle = await f.read()
        loaded = pickle.loads(serialized_pickle, encoding=encoding)
        if isinstance(loaded, dict):
            loaded.update(to_dump)
        elif isinstance(loaded, list):
            loaded += to_dump
        elif isinstance(loaded, np.ndarray):
            loaded = np.vstack((loaded, to_dump))
        else:
            raise ValueError
        serialized_pickle = pickle.dumps(loaded, protocol=2)
        async with aiofiles.open(file_name, mode='wb') as f:
            await f.write(serialized_pickle)
        print(file_name, 'dumped!')
    elif mode == 'w' or mode == 'a':
        serialized_pickle = pickle.dumps(to_dump, protocol=2)
        async with aiofiles.open(file_name, mode='wb') as f:
            await f.write(serialized_pickle)
        print(file_name, 'dumped!')
    else:
        raise ValueError


async def pickles_aio(file_names: Union[list, tuple], mode: str, to_dumps=None, key_option=None) -> dict:
    if mode == 'r':
        loaded = await asyncio.gather(*[asyncio.ensure_future(pickle_aio(file_name, mode=mode))
                                        for file_name in file_names])
        return {key_modifier(key, option=key_option): value for key, value in zip(file_names, loaded)}
    else:
        await asyncio.gather(*[asyncio.ensure_future(pickle_aio(file_name, mode=mode, to_dump=to_dump))
                               for to_dump, file_name in zip(to_dumps, file_names)])


def pickle_io(file_name: str, mode: str, to_dump: object = None) -> any:
    return asyncio.run(pickle_aio(file_name=file_name, mode=mode, to_dump=to_dump))


def pickles_io(file_names: Union[list, tuple], mode: str, to_dumps=None, key_option=None) -> dict:
    return asyncio.run(pickles_aio(file_names=file_names, mode=mode, to_dumps=to_dumps, key_option=key_option))


# def open_history_output(gen, path=None):
#     if path is not None:
#         from os import chdir
#         chdir(path)
#     from pickle import load
#     file_names = ('U1_HistoryOutput', 'U2_HistoryOutput', 'U3_HistoryOutput',
#                   'RF1_HistoryOutput', 'RF2_HistoryOutput', 'RF3_HistoryOutput')
#     for file_name in file_names:
#         pickle_file_name = file_name + f'_{gen}'
#         with open(pickle_file_name, 'rb') as f:
#             loaded_file = load(f, encoding='bytes')
#             print(f'Gen{gen}-{pickle_file_name}: {loaded_file}')



"""
Run full GA steps
"""


def run():
    """
    <!> Caution
    You must use if __name__ == '__main__' for running GA,
    If __name__=='__main__' is not used, multiple servers are running and multiple bindings occur on the same port,
    which causes OSError: [WinError 10048] server-binding error. The reason this phenomenon occurs is as follows.

    [1] Some functions like make_and_start_process() and 'start_abaqus_cae() creates new process.
    [2] If creating process using multiprocessing.Process(), that process will run the code at the top level of the script.
    [3] If you have code that you only want to run when the script is run directly, you can put it in the block of code
    indented under if __name__ == '__main__':, so that it will not be executed when the script is imported as a module and
    run in a new process. Without that condition, the same block of code will run multiple times in multiple processes
    which can lead to unexpected results.
    """
    import os
    from ..GraphicUserInterface import App, Visualizer, plot_previously_plotted_data
    from ..GeneticAlgorithm import NSGAModel
    from ..Network import Server, make_and_start_process, start_abaqus_cae
    from ..ParameterDefinitions import material_property_definitions
    import warnings
    warnings.warn(
        "You must use conditional block [if __name__ == '__main__'] for running GA. Otherwise, [WinError 10048] occurs.")

    HOST = 'localhost'
    PORT = 12345

    # Open socket server
    server = Server(host=HOST, port=PORT, option='json', run_nonblocking=True)

    try:
        # Make an interface and receive parameters
        gui_process, parent_conn, child_conn = make_and_start_process(target=App, duplex=True, daemon=True)
        set_path, parameters = parent_conn.recv()
        parameters.post_initialize()
        os.chdir(set_path)

        # Start ABAQUS
        abaqus_process = start_abaqus_cae(option=parameters.abaqus_mode)

        # Load previous plotting data
        visualizer = Visualizer(conn_to_gui=parent_conn)
        plot_previously_plotted_data(visualizer=visualizer, params=parameters)

        # Run GA process
        ga_model = NSGAModel(params=parameters, material_properties=material_property_definitions,
                             visualizer=visualizer, random_topology_density=0.3)
        ga_model.run(server=server)
    except Exception as e:
        print(e)
    finally:
        server.close()


if __name__ == '__main__':
    run()



import random
import numpy as np
import itertools
from functools import reduce
from dataclasses import asdict
from typing import Tuple
from .ParameterDefinitions import Parameters, JsonFormat, fitness_definitions
from .GraphicUserInterface import Visualizer
from .Network import Server, request_abaqus
from .FileIO import pickle_io, pickles_io, remove_file, get_sorted_file_numbers_from_pattern
from .PostProcessing import evaluate_all_fitness_values, selection
from .MutateAndValidate import mutate_and_validate_topology


class NSGAModel:
    def __init__(self, params: Parameters, material_properties: dict, visualizer: Visualizer = None,
                 random_topology_density: float = 0.5):
        self.params = params
        self.visualizer = visualizer
        self.material_properties = material_properties
        self.random_topology_density = random_topology_density

    def load_parent_data(self, gen: int, server: Server) -> Tuple[np.ndarray, dict]:
        try:
            parent_topologies = pickle_io(f'Topologies_{gen}', mode='r')['parent']
        except FileNotFoundError or KeyError:
            parent_topologies = random_parent_generation(density=self.random_topology_density,
                                                         params=self.params, save_file_as=f'Topologies_{gen}')
        try:
            parent_results = pickle_io(f'FieldOutput_{gen}', mode='r')
        except FileNotFoundError:
            json_data = asdict(JsonFormat(start_topology_from=1, topologies_key='parent',
                                          topologies_file_name=f'Topologies_{gen}', exit_abaqus=False))
            json_data.update(asdict(self.params))
            json_data.update(self.material_properties)
            request_abaqus(dict_data=json_data, server=server)
            parent_results = pickle_io(f'FieldOutput_offspring_{gen}', mode='r')
            remove_file(f'FieldOutput_offspring_{gen}')
            pickle_io(f'FieldOutput_{gen}', mode='w', to_dump=parent_results)
            self.visualizer.visualize(params=self.params, gen=gen - 1, use_manual_rp=False)
        assert len(parent_topologies) == len(parent_results)
        return parent_topologies, parent_results

    def determine_where_abaqus_start(self) -> Tuple[int, int]:
        offspring_results_file_numbers = get_sorted_file_numbers_from_pattern(r'FieldOutput_offspring_\d+')
        if len(offspring_results_file_numbers) == 0:
            return 1, 1
        else:
            last_offspring_results_file_number = offspring_results_file_numbers[-1]
            last_offspring_results = pickle_io(f'FieldOutput_offspring_{last_offspring_results_file_number}', mode='r')
            if len(last_offspring_results) < self.params.end_pop:
                needed_to_start_at_offspring = len(last_offspring_results) + 1
                return last_offspring_results_file_number, needed_to_start_at_offspring
            else:
                return last_offspring_results_file_number + 1, 1

    def generate_offspring_topologies(self, gen: int, server: Server) -> np.ndarray:
        parent_topologies, parent_results = self.load_parent_data(gen=gen, server=server)
        topologies = pickle_io(f'Topologies_{gen}', mode='r')
        if 'offspring' in topologies.keys():
            offspring_topologies = pickle_io(f'Topologies_{gen}', mode='r')['offspring']
        else:
            offspring_topologies = generate_offspring(gen=gen, topo_parents=parent_topologies, params=self.params,
                                                      save_file_as=f'Topologies_{gen}')
        assert len(parent_topologies) == len(parent_results) == len(offspring_topologies)
        return offspring_topologies

    def run_a_generation(self, running_gen: int, start_offspring_from: int, server: Server):
        parent_topologies, parent_results = self.load_parent_data(gen=running_gen, server=server)
        offspring_topologies = self.generate_offspring_topologies(gen=running_gen, server=server)
        json_data = asdict(JsonFormat(start_topology_from=start_offspring_from, topologies_key='offspring',
                                      topologies_file_name=f'Topologies_{running_gen}', exit_abaqus=False))
        json_data.update(asdict(self.params))
        json_data.update(self.material_properties)
        request_abaqus(dict_data=json_data, server=server)
        offspring_results = pickle_io(f'FieldOutput_offspring_{running_gen}', mode='r')
        all_topologies = np.vstack((parent_topologies, offspring_topologies))
        all_results = parent_results.copy()
        all_results.update({entity_num + len(parent_results): offspring_results[entity_num]
                            for entity_num in sorted(offspring_results.keys())})
        all_fitness_values = evaluate_all_fitness_values(fitness_definitions=fitness_definitions,
                                                         params_dict=asdict(self.params),
                                                         results=all_results, topologies=all_topologies)
        pareto_indices = selection(all_fitness_values=all_fitness_values, selected_size=self.params.end_pop)
        selected_topologies = all_topologies[pareto_indices]
        selected_results = {entity_num: all_results[pareto_idx + 1]
                            for entity_num, pareto_idx in enumerate(pareto_indices, start=1)}
        assert len(selected_topologies) == len(selected_results)
        pickle_io(f'Topologies_{running_gen + 1}', mode='w', to_dump={'parent': selected_topologies})
        pickle_io(f'FieldOutput_{running_gen + 1}', mode='w', to_dump=selected_results)
        if self.visualizer is not None:
            self.visualizer.visualize(params=self.params, gen=running_gen, use_manual_rp=False)

    def run(self, server):
        start_gen, start_offspring = self.determine_where_abaqus_start()
        for gen in range(start_gen, self.params.end_gen):
            self.run_a_generation(running_gen=gen, start_offspring_from=start_offspring, server=server)
            start_offspring = 1
        request_abaqus(dict_data={'exit_abaqus': True}, server=server)


def find_where_same_array_locates(arr_to_find: np.ndarray, big_arr: np.ndarray) -> np.ndarray:
    axis_range = tuple(dim for dim in range(len(big_arr.shape) - len(arr_to_find.shape), len(big_arr.shape)))
    return np.argwhere(np.all(big_arr == arr_to_find, axis=axis_range))


def get_cutting_section_and_candidates(topologies: np.ndarray) -> Tuple[int, np.ndarray]:
    topologies_flattened = topologies.reshape((len(topologies), -1))
    possible_cutting_sections = np.random.permutation(topologies_flattened.shape[1])
    for possible_cutting_section in possible_cutting_sections:
        candidates_for_cutting = np.empty((0,), dtype=int)
        for topo_idx in range(len(topologies)):
            if topologies_flattened[topo_idx, possible_cutting_section] == 1:
                candidates_for_cutting = np.hstack((candidates_for_cutting, topo_idx))
        if len(candidates_for_cutting) >= 2:
            return possible_cutting_section, candidates_for_cutting
    raise RuntimeError


def get_candidate_pairs(candidates):
    candidates_results = list(itertools.combinations(candidates, 2))  # possible candidate pair of parents
    random.shuffle(candidates_results)
    return candidates_results


def crossover(chromosome_1, chromosome_2, cutting_section):  # Crossover process
    chromosome_1_flattened = chromosome_1.flatten()
    chromosome_2_flattened = chromosome_2.flatten()
    cross_overed_chromosome_1 = np.empty_like(chromosome_1_flattened)
    cross_overed_chromosome_2 = np.empty_like(chromosome_2_flattened)
    cross_overed_chromosome_1[0:cutting_section] = chromosome_1_flattened[0:cutting_section]
    cross_overed_chromosome_1[cutting_section:] = chromosome_2_flattened[cutting_section:]
    cross_overed_chromosome_2[0:cutting_section] = chromosome_2_flattened[0:cutting_section]
    cross_overed_chromosome_2[cutting_section:] = chromosome_1_flattened[cutting_section:]
    return cross_overed_chromosome_1.reshape(chromosome_1.shape), cross_overed_chromosome_2.reshape(chromosome_2.shape)


def generate_offspring(gen: int, params: Parameters, topo_parents: np.ndarray, save_file_as: str = None) -> np.ndarray:
    """
    Generating topo_offspring from parents. Crossover & Mutation & Validating processes will be held.
    Validating processes contain, checking 3d-print-ability without support, one voxel tree contacting six faces
    in a cube
    :param save_file_as: Save pickle file of offspring topologies as this.
    :param topo_parents: Topology array of parent, shape: (end_pop, lx, ly, lz)
    :param gen: Current generation
    :param params: Parameters for GA
    The more timeout, the longer time will be allowed for the function "mutate_and_validate_topology".
    :return:
    """
    topo_offspring = np.empty((0, params.lx, params.ly, params.lz), int)
    validation_count = 0
    all_topos = pickles_io(file_names=[f'Topologies_{g}' for g in range(1, gen + 1)], mode='r', key_option='int')
    all_topos_parent = np.array([topos['parent'] for topos in all_topos.values()], dtype=int)
    while True:
        cutting_section, candidates = get_cutting_section_and_candidates(topologies=topo_parents)
        candidate_pairs = get_candidate_pairs(candidates=candidates)
        print('<info> Candidate lists:', len(candidates))
        print('<info> Candidate pairs:', len(candidate_pairs))
        for pair_idx in range(len(candidate_pairs)):
            chromosome_1_idx = candidate_pairs[pair_idx][0]
            chromosome_2_idx = candidate_pairs[pair_idx][1]
            cross_overed_chromosome_1, cross_overed_chromosome_2 = crossover(
                chromosome_1=topo_parents[chromosome_1_idx],
                chromosome_2=topo_parents[chromosome_2_idx],
                cutting_section=cutting_section)
            validated_chromosome_1 = mutate_and_validate_topology(cross_overed_chromosome_1, params.mutation_rate)
            validated_chromosome_2 = mutate_and_validate_topology(cross_overed_chromosome_2, params.mutation_rate)
            for validated_chromosome in (validated_chromosome_1, validated_chromosome_2):
                if validated_chromosome is None:
                    print('<!> Non-connected tree detected')
                    continue
                else:
                    if len(find_where_same_array_locates(arr_to_find=validated_chromosome,
                                                         big_arr=all_topos_parent)):
                        print('<!> Clone structure found in parents!')
                        continue
                    elif len(find_where_same_array_locates(arr_to_find=validated_chromosome,
                                                           big_arr=topo_offspring)):
                        print('<!> Clone structure found in current offsprings!')
                        continue
                    else:
                        topo_offspring = np.vstack((topo_offspring, np.expand_dims(validated_chromosome, axis=0)))
                        validation_count += 1
                        print(f'<info> Validation of chromosome {validation_count} complete!')
                        if len(topo_offspring) == params.end_pop:
                            print('<info> Generating topo_offspring complete')
                            if save_file_as is not None:
                                pickle_io(save_file_as, mode='a', to_dump={'offspring': topo_offspring})
                            return topo_offspring


def random_array(shape, probability):
    return np.random.choice([1, 0], size=reduce(lambda x, y: x * y, shape), p=[probability, 1 - probability]).reshape(
        shape)


def random_parent_generation(density: float, params: Parameters, save_file_as: str = None) -> np.ndarray:
    parents = np.empty((params.end_pop, params.lx, params.ly, params.lz))
    total_parent_generation_count = 0
    total_volume_frac = 0
    while total_parent_generation_count < params.end_pop:
        print(f'<<<<< Parent {total_parent_generation_count + 1} >>>>>')
        while True:
            rand_arr = random_array(shape=(params.lx, params.ly, params.lz), probability=density)
            parent = mutate_and_validate_topology(rand_arr, mutation_probability=params.mutation_rate)
            if parent is not None:
                break
        volume_frac = np.count_nonzero(parent) / (params.lx * params.ly * params.lz / 100)
        total_volume_frac += volume_frac
        print(f'Volume fraction: {volume_frac:.1f} %\n')
        parents[total_parent_generation_count] = parent
        total_parent_generation_count += 1
    print(f'Average volume fraction: {total_volume_frac / params.end_pop:.1f} %')
    if save_file_as is not None:
        pickle_io(save_file_as, mode='w', to_dump={'parent': parents})
    return parents



import regionToolset
from abaqus import *
from abaqusConstants import *
from driverUtils import executeOnCaeStartup
from odbAccess import openOdb
import numpy as np
import pickle
import os
from datetime import datetime
import threading
import socket
import struct
import json
from sys import version_info
try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk
try:
    from Queue import Queue
except ImportError:
    from queue import Queue

executeOnCaeStartup()
HOST = 'localhost'
PORT = 12345


class Client:
    def __init__(self, host, port, option, connect):
        self.host = host
        self.port = port
        self.option = option
        self.q = Queue()
        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.is_alive = True
        self._default_packet_size = 1024
        self._header_format = '>I'
        self._header_bytes = 4
        if connect:
            self.connect()

    def connect(self):
        if version_info.major >= 3:
            new_th = threading.Thread(target=self._thread_recv, args=(self.client_socket, self.option), daemon=True)
        else:
            new_th = threading.Thread(target=self._thread_recv, args=(self.client_socket, self.option))
            new_th.setDaemon(True)
        self.client_socket.connect((self.host, self.port))
        print('[{}] Connected to {}:{}'.format(datetime.now(), self.host, self.port))
        new_th.start()

    def send(self, data):
        if self.option == 'pickle':
            serialized_data = pickle.dumps(data, protocol=2)
        else:
            serialized_data = json.dumps(data).encode()
        while True:
            try:
                self.client_socket.sendall(struct.pack(self._header_format, len(serialized_data)))
                self.client_socket.sendall(serialized_data)
                print('[{}] A data sent'.format(datetime.now()))
                break
            except Exception as send_error:
                print('Sending data failed, trying to reconnect to server: ', send_error)
                self.connect()

    def recv(self):
        return self.q.get()

    def close(self):
        self.client_socket.close()

    def _thread_recv(self, client_socket, option):
        while True:
            try:  # Trying to receive a data and decode it
                data_size = struct.unpack(self._header_format, client_socket.recv(self._header_bytes))[0]
                remaining_payload_size = data_size
                packets = b''
                while remaining_payload_size != 0:
                    packets += client_socket.recv(remaining_payload_size)
                    remaining_payload_size = data_size - len(packets)
                try:  # Trying to decode received data
                    if option == 'json':
                        received_data = json.loads(packets.decode())
                    else:
                        if version_info.major >= 3:
                            received_data = pickle.loads(packets, encoding='bytes')
                        else:
                            received_data = pickle.loads(packets)
                    print('[{}] Received data: {}'.format(datetime.now(), received_data))
                    self.q.put(received_data)
                except Exception as e2:  # Decoding is failed
                    print('[{}] Loading received data failure: {}'.format(datetime.now(), e2))
                    continue
            except Exception as e1:  # Connection is lost
                print('[{}] Error: {}'.format(datetime.now(), e1))
                break
        self.is_alive = False
        print('<!> Connection dead')


class JobLogFrame(tk.Frame):
    def __init__(self, *args, **kwargs):
        tk.Frame.__init__(self, *args, **kwargs)
        self.text = tk.Text(self, height=50, width=100)
        self.vsb = tk.Scrollbar(self, orient="vertical", command=self.text.yview)
        self.text.configure(yscrollcommand=self.vsb.set)
        self.vsb.pack(side="right", fill="y")
        self.text.pack(side="left", fill="both", expand=True)


class MyModel:
    def __init__(self, model_name, params):
        session.journalOptions.setValues(replayGeometry=COORDINATE, recoverGeometry=COORDINATE)
        self.model = mdb.Model(name=model_name, modelType=STANDARD_EXPLICIT)
        self.root_assembly = self.model.rootAssembly
        self.root_assembly.DatumCsysByDefault(CARTESIAN)
        self.params = params

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if len(mdb.models.keys()) == 1:
            mdb.Model(name='empty_model', modelType=STANDARD_EXPLICIT)
        del mdb.models[self.model.name]

    def create_voxel_part(self, voxel_name):
        _voxel_part = self.model.Part(name=voxel_name, dimensionality=THREE_D, type=DEFORMABLE_BODY)
        _s = self.model.ConstrainedSketch(name='__profile__', sheetSize=2 * self.params['unit_l'])
        _s.rectangle(point1=(0.0, 0.0), point2=(self.params['unit_l'], self.params['unit_l']))
        _voxel_part.BaseSolidExtrude(sketch=_s, depth=self.params['unit_l'])

    def create_cube_part(self, voxel_name, cube_name, topo_arr):
        for ix in range(topo_arr.shape[0]):
            for iy in range(topo_arr.shape[1]):
                for iz in range(topo_arr.shape[2]):
                    if topo_arr[ix, iy, iz]:
                        _instance_name = '{}-{}-{}-{}'.format(cube_name, ix, iy, iz)
                        self.root_assembly.Instance(name=_instance_name, part=self.model.parts[voxel_name],
                                                    dependent=ON)
                        self.root_assembly.translate(instanceList=(_instance_name,),
                                                     vector=(ix * self.params['unit_l'],
                                                             iy * self.params['unit_l'],
                                                             iz * self.params['unit_l']))
        self.merge_mesh_of_instances(part_name_of_merged_assembly=cube_name)

    def create_mesh_of_part(self, part_name):
        self.model.parts[part_name].seedPart(size=self.params['mesh_size'], minSizeFactor=0.9, deviationFactor=0.1)
        self.model.parts[part_name].generateMesh()

    def merge_mesh_of_instances(self, part_name_of_merged_assembly):
        _instances = self.root_assembly.instances.values()
        self.root_assembly.InstanceFromBooleanMerge(name=part_name_of_merged_assembly, instances=_instances,
                                                    originalInstances=DELETE, mergeNodes=BOUNDARY_ONLY,
                                                    nodeMergingTolerance=1e-06, domain=MESH)

    def create_material(self, material_name, density, engineering_constants):

        self.model.Material(name=material_name)
        self.model.materials[material_name].Density(table=((density,),))
        self.model.materials[material_name].Elastic(type=ENGINEERING_CONSTANTS, table=(engineering_constants,))

    def assign_section_to_elements_of_part_by_bounding_box(self, part_name, material_name, section_name,
                                                           bound_definition):
        bounded_elements = self.model.parts[part_name].elements.getByBoundingBox(**bound_definition)
        bounded_region = regionToolset.Region(elements=bounded_elements)
        self.model.HomogeneousSolidSection(material=material_name, name=section_name, thickness=None)
        self.model.parts[part_name].SectionAssignment(region=bounded_region, sectionName=section_name,
                                                      offset=0.0, offsetType=MIDDLE_SURFACE, offsetField='',
                                                      thicknessAssignment=FROM_SECTION)
        self.model.parts[part_name].MaterialOrientation(additionalRotationType=ROTATION_NONE, axis=AXIS_1,
                                                        fieldName='', localCsys=None, orientationType=GLOBAL,
                                                        region=bounded_region, stackDirection=STACK_3)

    def create_set_of_part_by_bounding_box(self, part_name, set_name, bound_definition):
        self.model.parts[part_name].Set(name=set_name,
                                        nodes=self.model.parts[part_name].nodes.getByBoundingBox(**bound_definition))

    def create_set_by_bounding_box(self, instance_name, set_name, bound_definition):
        self.root_assembly.Set(name=set_name, nodes=self.root_assembly.instances[instance_name].nodes.getByBoundingBox(
            **bound_definition))

    def set_encastre(self, bc_name, set_name, step_name):
        self.model.EncastreBC(name=bc_name, createStepName=step_name,
                              localCsys=None, region=self.root_assembly.sets[set_name])

    def set_displacement(self, bc_name, set_name, step_name, displacement):
        self.model.DisplacementBC(name=bc_name, createStepName=step_name,
                                  amplitude=UNSET, distributionType=UNIFORM, fieldName='', localCsys=None,
                                  region=self.root_assembly.sets[set_name], **displacement)

    def set_boundary_condition(self, symmetry_direction, set_name):
        if symmetry_direction == 'x':
            self.model.XsymmBC(createStepName='Initial', localCsys=None, name='x_sym',
                               region=self.root_assembly.sets[set_name])
        elif symmetry_direction == 'y':
            self.model.YsymmBC(createStepName='Initial', localCsys=None, name='y_sym',
                               region=self.root_assembly.sets[set_name])
        elif symmetry_direction == 'z':
            self.model.ZsymmBC(createStepName='Initial', localCsys=None, name='z_sym',
                               region=self.root_assembly.sets[set_name])
        else:
            raise ValueError

    def create_step(self, step_name, previous_step, step_type):
        if step_type == 'modal':
            self.model.FrequencyStep(name=step_name, previous=previous_step,
                                     limitSavedEigenvectorRegion=None, numEigen=12)
        elif step_type == 'compression':
            self.model.StaticStep(initialInc=0.001, maxInc=0.1, maxNumInc=10000, minInc=1e-12,
                                  name=step_name, previous=previous_step)
        else:
            raise ValueError

    def create_output_requests(self, step_name, history_output_name, set_name,
                               field_outputs, history_outputs):
        self.model.fieldOutputRequests['F-Output-1'].setValues(variables=field_outputs)
        self.model.HistoryOutputRequest(createStepName=step_name, name=history_output_name, rebar=EXCLUDE,
                                        region=self.root_assembly.sets[set_name], sectionPoints=DEFAULT,
                                        variables=history_outputs)

    def create_reference_point_and_set(self, rp_name, rp_coordinate):
        rp_id = self.root_assembly.ReferencePoint(point=rp_coordinate).id
        self.root_assembly.Set(name=rp_name, referencePoints=(self.root_assembly.referencePoints[rp_id],))
        return rp_id

    def create_coupling(self, rp_set_name, surface_set_name, constraint_name):
        self.model.Coupling(alpha=0.0, controlPoint=self.root_assembly.sets[rp_set_name], couplingType=KINEMATIC,
                            influenceRadius=WHOLE_SURFACE, localCsys=None, name=constraint_name,
                            surface=self.root_assembly.sets[surface_set_name],
                            u1=ON, u2=ON, u3=ON, ur1=ON, ur2=ON, ur3=ON)

    def allow_self_contact(self, instance_name, step_name):
        elements = self.root_assembly.instances[instance_name].elements.getExteriorFaces()
        _surface = self.root_assembly.Surface(name='ExteriorSurface', side1Elements=elements)
        _interaction_property_name = 'SelfContactProp'
        _interaction_name = 'SelfContact'
        self.model.ContactProperty(_interaction_property_name)
        self.model.interactionProperties[_interaction_property_name].NormalBehavior(
            allowSeparation=ON, constraintEnforcementMethod=DEFAULT, pressureOverclosure=HARD)
        self.model.SelfContactStd(name=_interaction_name, contactTracking=ONE_CONFIG, createStepName=step_name,
                                  interactionProperty=_interaction_property_name, surface=_surface, thickness=ON)

    def create_job(self, job_name, num_cpus, num_gpus, run):
        mdb.Job(name=job_name, model=self.model.name, description='',
                type=ANALYSIS, atTime=None, waitMinutes=0, waitHours=0, queue=None, memory=90,
                memoryUnits=PERCENTAGE, getMemoryFromAnalysis=True,
                explicitPrecision=SINGLE, nodalOutputPrecision=SINGLE, echoPrint=OFF,
                modelPrint=OFF, contactPrint=OFF, historyPrint=OFF, userSubroutine='',
                scratch='', resultsFormat=ODB, numThreadsPerMpiProcess=1,
                multiprocessingMode=DEFAULT, numCpus=num_cpus, numGPUs=num_gpus)
        if run:
            mdb.jobs[job_name].submit(consistencyChecking=OFF)
            mdb.jobs[job_name].waitForCompletion()


# Function for Python 2
def ascii_encode_dict(data):
    ascii_encode = lambda x: x.encode('ascii') if isinstance(x, unicode) else x
    return dict(map(ascii_encode, pair) for pair in data.items())


def open_job_log():
    _root = tk.Tk()
    _root.title('Abaqus control log')
    _frame = JobLogFrame(_root)
    _frame.pack(fill="both", expand=True)
    _new_thread = threading.Thread(target=_root.mainloop)
    _new_thread.setDaemon(True)
    _new_thread.start()
    return _frame


def save_log(message, job_log_frame):
    _now = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
    _message = '[{}] {}\n'.format(_now, message)
    print(_message)
    with open('log.txt', mode='a') as f_log:
        f_log.write(_message)
    job_log_frame.text.insert('end', _message)
    job_log_frame.text.see('end')


def dump_pickled_dict_data(file_name, key, to_dump, mode):
    if mode == 'a' and os.path.isfile(file_name):
        with open(file_name, mode='rb') as f:
            dict_data = pickle.load(f)
        dict_data.update({key: to_dump})
    else:
        dict_data = {key: to_dump}
    with open(file_name, mode='wb') as f:
        pickle.dump(dict_data, f)


def load_pickled_dict_data(file_name):
    with open(file_name, mode='rb') as f:
        dict_data = pickle.load(f)
    return dict_data


def random_array(shape, probability):
    from functools import reduce
    return np.random.choice([1, 0], size=reduce(lambda x, y: x * y, shape),
                            p=[probability, 1 - probability]).reshape(shape)


def quaver_to_full(quaver):
    quarter = np.concatenate((np.flip(quaver, axis=0), quaver), axis=0)
    half = np.concatenate((np.flip(quarter, axis=1), quarter), axis=1)
    full = np.concatenate((np.flip(half, axis=2), half), axis=2)
    return np.swapaxes(full, axis1=0, axis2=2)


def bound_setter(whole_bound, option):
    """
    Create dictionary of bound limits used for bounding box
    :param whole_bound: whole bound dictionary for bounding box, six keys in dictionary
    :param option: This can be either 'xMin', 'yMin', 'zMin', 'xMax', 'yMax', 'zMax'. If 'xMin',
     bound for x=0 will be returned.
    :return: a dictionary for bounding box
    """
    bound = whole_bound.copy()
    if 'Min' in option:
        bound[option[0] + 'Max'] = bound[option[0] + 'Min']
    elif 'Max' in option:
        bound[option[0] + 'Min'] = bound[option[0] + 'Max']
    else:
        pass
    return bound


def export_outputs(model_name, step_name, rp_name):
    gen, entity = map(int, model_name.split('-'))
    history_output_file_header = 'HistoryOutput_offspring'
    field_output_file_header = 'FieldOutput_offspring'
    exported_field_outputs = {
        'displacement': {'xMax': np.ndarray, 'yMax': np.ndarray, 'zMax': np.ndarray},
        'rotation': np.ndarray,
        'reaction_force': np.ndarray,
        'mises_stress': {'max': float, 'min': float, 'average': float}
    }
    exported_history_output_properties = ('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3')
    try:
        odb = openOdb('Job-{}.odb'.format(model_name))
        try:
            # Field outputs
            field_outputs = {
                'displacement': odb.steps[step_name].frames[-1].fieldOutputs['U'],
                'rotation': odb.steps[step_name].frames[-1].fieldOutputs['UR'],
                'reaction_force': odb.steps[step_name].frames[-1].fieldOutputs['RF'],
                'stress': odb.steps[step_name].frames[-1].fieldOutputs['S']
            }
            displacement_of_node_sets = {
                face: field_outputs['displacement'].getSubset(region=odb.rootAssembly.nodeSets[face.upper()]).values
                for face in ('xMax', 'yMax', 'zMax')
            }
            for face, displacement_of_node_set in displacement_of_node_sets.items():
                exported_field_outputs['displacement'][face] = np.average(
                    np.array([data.data for data in displacement_of_node_set]), axis=0)
            exported_field_outputs['reaction_force'] = field_outputs['reaction_force'].getSubset(
                region=odb.rootAssembly.nodeSets[rp_name.upper()]).values[0].data
            exported_field_outputs['rotation'] = field_outputs['rotation'].getSubset(
                region=odb.rootAssembly.nodeSets[rp_name.upper()]).values[0].data
            _values = [stress.mises for stress in field_outputs['stress'].values]
            exported_field_outputs['mises_stress']['max'] = max(_values)
            exported_field_outputs['mises_stress']['min'] = min(_values)
            exported_field_outputs['mises_stress']['average'] = np.average(_values)
            dump_pickled_dict_data(file_name='{}_{}'.format(field_output_file_header, str(gen)),
                                   key=entity, to_dump=exported_field_outputs, mode='a')

            # History outputs
            history_outputs = odb.steps[step_name].historyRegions['Node ASSEMBLY.1'].historyOutputs
            exported_history_outputs = {prop: np.array(history_outputs[prop].data)
                                        for prop in exported_history_output_properties}
            dump_pickled_dict_data(file_name='{}_{}'.format(history_output_file_header, str(gen)),
                                   key=entity, to_dump=exported_history_outputs, mode='a')

        except Exception as e2:
            print('ODB output reading failed: ', e2)
        finally:
            odb.close()
    except Exception as e1:
        print('There is no ODB file: ', e1)


def run_analysis(params, model_name, topo_arr, voxel_name, voxel_unit_length, cube_name,
                 analysis_mode, material_properties, full, displacement=None):
    topo_arr = quaver_to_full(topo_arr) if full else topo_arr.copy()
    cube_x_voxels, cube_y_voxels, cube_z_voxels = topo_arr.shape
    cube_x_size = voxel_unit_length * cube_x_voxels
    cube_y_size = voxel_unit_length * cube_y_voxels
    cube_z_size = voxel_unit_length * cube_z_voxels
    whole_bound = {'xMin': 0., 'yMin': 0., 'zMin': 0., 'xMax': cube_x_size, 'yMax': cube_y_size, 'zMax': cube_z_size}
    bounds = {option: bound_setter(whole_bound=whole_bound, option=option) for option in whole_bound.keys()}
    rp_coordinates = {'RP-x': (1.05 * cube_x_size, cube_y_size / 2, cube_z_size / 2),
                      'RP-y': (cube_x_size / 2, 1.05 * cube_y_size, cube_z_size / 2),
                      'RP-z': (cube_x_size / 2, cube_y_size / 2, 1.05 * cube_z_size)}

    with MyModel(model_name='Model-{}'.format(model_name), params=params) as mm:
        material_name = material_properties['material_name']
        mm.create_voxel_part(voxel_name=voxel_name)
        mm.create_mesh_of_part(part_name=voxel_name)
        mm.create_cube_part(voxel_name=voxel_name, cube_name=cube_name, topo_arr=topo_arr)
        mm.create_material(**material_properties)
        mm.assign_section_to_elements_of_part_by_bounding_box(part_name=cube_name, material_name=material_name,
                                                              section_name=material_name + '-section',
                                                              bound_definition=whole_bound)
        for option, bound in bounds.items():
            mm.create_set_by_bounding_box(instance_name=cube_name + '-1', set_name=option, bound_definition=bound)
        for rp_name, rp_coordinate in rp_coordinates.items():
            mm.create_reference_point_and_set(rp_name=rp_name, rp_coordinate=rp_coordinate)
        for symmetry_diction, boundary_set_name in (('x', 'xMin'), ('y', 'yMin'), ('z', 'zMin')):
            mm.set_boundary_condition(symmetry_direction=symmetry_diction, set_name=boundary_set_name)
        if analysis_mode == 'modal':
            mm.create_step(step_name=analysis_mode + '-step', previous_step='Initial', step_type=analysis_mode)
            mm.set_encastre(bc_name='encastre_bottom', set_name='yMin', step_name='Initial')
        elif analysis_mode == 'compression':
            analysis_step_name = analysis_mode + '-step'
            mm.create_step(step_name=analysis_step_name, previous_step='Initial', step_type=analysis_mode)
            mm.allow_self_contact(instance_name=cube_name + '-1', step_name=analysis_step_name)
            mm.create_coupling(rp_set_name='RP-y', surface_set_name='yMax', constraint_name='coupling')
            mm.set_displacement(bc_name='displacement',
                                set_name='RP-y',
                                step_name=analysis_step_name, displacement=displacement)
            mm.create_output_requests(step_name=analysis_step_name, history_output_name='H-Output', set_name='RP-y',
                                      field_outputs=('S', 'U', 'RF', 'IVOL', 'MISESMAX'),
                                      history_outputs=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'ALLIE'))
        else:
            raise ValueError
        mm.root_assembly.regenerate()
        mm.create_job(job_name='Job-{}'.format(model_name),
                      num_cpus=params['n_cpus'], num_gpus=params['n_gpus'], run=True)
        export_outputs(model_name=model_name, step_name=analysis_step_name, rp_name='RP-y')


if __name__ == '__main__':
    client = Client(host=HOST, port=PORT, option='json', connect=True)
    frame = open_job_log()
    save_log('Connected to {}:{}'.format(PORT, HOST), job_log_frame=frame)
    while True:
        parameters = client.recv()
        parameters = ascii_encode_dict(parameters)
        if parameters['exit_abaqus']:
            break
        material_property_definitions = {
            'material_name': parameters['material_name'],
            'density': parameters['density'],
            'engineering_constants': parameters['engineering_constants']
        }
        start_topology_from = parameters['start_topology_from']
        topologies_file_name = parameters['topologies_file_name']
        topologies_key = parameters['topologies_key']
        topologies = load_pickled_dict_data(topologies_file_name)[topologies_key]
        gen_num = topologies_file_name.split('_')[-1]
        for entity_num, topology in enumerate(topologies, start=1):
            if entity_num < start_topology_from:
                continue
            run_analysis(model_name='{}-{}'.format(gen_num, entity_num), analysis_mode='compression',
                         topo_arr=topology, voxel_unit_length=parameters['unit_l'], full=False, params=parameters,
                         material_properties=material_property_definitions, voxel_name='voxel', cube_name='cube',
                         displacement={'u1': 0, 'u2': parameters['dis_y'], 'u3': 0, 'ur1': 0, 'ur2': 0, 'ur3': 0})
            save_log('Created Job{}-{}.odb'.format(gen_num, entity_num), job_log_frame=frame)
        client.send('[{}] Generation {} finished!'.format(datetime.now(), gen_num))



from . import FileIO
from . import GeneticAlgorithm
from . import GraphicUserInterface
from . import MutateAndValidate
from . import Network
from . import PostProcessing
from . import ParameterDefinitions
from .GeneticAlgorithm import *
from .FileIO import *
from .GraphicUserInterface import *
from .MutateAndValidate import *
from .Network import *
from .PostProcessing import *
from .ParameterDefinitions import *



import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter
from typing import Union


def get_datum_hv(pareto_1_sorted: np.ndarray, pareto_2_sorted: np.ndarray) -> float:
    datum_point_x, datum_point_y = pareto_1_sorted[-1], pareto_2_sorted[0]
    xrs = pareto_1_sorted - datum_point_x
    yrs = pareto_2_sorted - datum_point_y
    datum_hv = 0
    for point_idx, (xr, yr) in enumerate(zip(xrs, yrs)):
        if point_idx == len(xrs) - 1:
            break
        datum_hv += abs((xrs[point_idx + 1] - xr) * (yrs[point_idx + 1] + yr) * 0.5)
    datum_hv -= abs((pareto_1_sorted[-1] - pareto_1_sorted[0]) * (pareto_2_sorted[0] - pareto_2_sorted[-1]))
    return datum_hv


def get_hv_from_datum_hv(datum_hv: float, lower_bounds: list, ref_x: float, ref_y: float):
    return datum_hv + (ref_x - lower_bounds[0]) * (ref_y - lower_bounds[1])


def evaluate_fitness_value_for_one_entity(vars_definitions: dict, fitness_value_definitions: Union[tuple, list],
                                          params_dict: dict, result: dict, topology: np.ndarray) -> np.ndarray:
    vars_dict = dict()
    predefined_vars = {
        'total_voxels': np.sum(topology),
        'max_rf22': params_dict['unit_l'] * params_dict['lx'] * params_dict['unit_l'] * params_dict['lz'] * params_dict['material_modulus']
    }
    for var, definition in vars_definitions.items():
        if isinstance(definition, (list, tuple, set)):
            _results = result.copy()
            for result_key in definition:
                _results = _results[result_key]
            vars_dict.update({var: _results})
        elif isinstance(definition, str):
            if definition.startswith('@'):
                vars_dict.update({var: params_dict[definition[1:]]})
            elif definition.startswith('$'):
                vars_dict.update({var: predefined_vars[definition[1:]]})
            else:
                raise ValueError
        else:
            raise ValueError
    locals().update(vars_dict)
    fitness_values = np.empty((1, len(fitness_value_definitions)), dtype=float)
    for fitness_value_idx, definition in enumerate(fitness_value_definitions):
        fitness_values[0, fitness_value_idx] = eval(definition)
    return fitness_values


def evaluate_all_fitness_values(fitness_definitions: dict, params_dict: dict,
                                results: dict, topologies: np.ndarray) -> np.ndarray:
    evaluation_version = params_dict['evaluation_version']
    vars_definitions = fitness_definitions[evaluation_version].vars_definitions
    fitness_value_definitions = fitness_definitions[evaluation_version].fitness_value_definitions

    all_fitness_values = np.empty((0, len(fitness_value_definitions)), dtype=float)
    assert len(topologies) == len(results)
    for entity_idx in range(len(topologies)):
        all_fitness_values = np.vstack((all_fitness_values, evaluate_fitness_value_for_one_entity(
            vars_definitions=vars_definitions, fitness_value_definitions=fitness_value_definitions,
            params_dict=params_dict, result=results[entity_idx+1], topology=topologies[entity_idx])))
    return all_fitness_values


def find_pareto_front_points(costs: np.ndarray, return_index: bool = False) -> np.ndarray:
    """
    A function calculating indices of pareto fronts or values of pareto points. The returned value is sorted by the
    first cost value.
    :param costs: The array containing fitness values, shape: (no_of_points x no_of_costs)
    :param return_index: If True, returns indices of pareto fronts. If False, returns fitness values of pareto fronts.
    :return: if return_index == True, The list of indices of pareto front points from costs, len: no_of_pareto_points
    if return_index == False, The array containing fitness values of pareto points, shape: (no_of_pareto_front_points
    x no_of_costs)
    """
    costs_copy = costs.copy()
    unsorted_indices = np.arange(len(costs_copy))
    next_point_idx = 0
    while next_point_idx < len(costs_copy):
        not_dominated_point_mask = np.any(costs_copy < costs_copy[next_point_idx], axis=1)
        not_dominated_point_mask[next_point_idx] = True
        unsorted_indices = unsorted_indices[not_dominated_point_mask]
        costs_copy = costs_copy[not_dominated_point_mask]
        next_point_idx = np.sum(not_dominated_point_mask[:next_point_idx]) + 1
    sorted_indices = unsorted_indices[np.argsort(costs[unsorted_indices][:, 0])]
    sorted_pareto_points = costs[sorted_indices]
    if return_index:
        return sorted_indices
    else:
        return sorted_pareto_points


def crowding_calculation(fitness_values: np.ndarray):
    population_size, number_of_fitness_values = fitness_values.shape[0], fitness_values.shape[1]
    matrix_for_crowding = np.zeros((population_size, number_of_fitness_values))
    normalize_fitness_values = (fitness_values - np.min(fitness_values, axis=0)) / np.ptp(fitness_values, axis=0)

    # normalize the fitness values
    for i in range(number_of_fitness_values):
        crowding_results = np.zeros(population_size)
        crowding_results[0] = 1  # extreme point has the max crowding distance
        crowding_results[population_size - 1] = 1  # extreme point has the max crowding distance
        sorting_normalize_fitness_values = np.sort(normalize_fitness_values[:, i])
        sorting_normalized_values_index = np.argsort(normalize_fitness_values[:, i])

        # crowding distance calculation
        crowding_results[1:population_size - 1] = (sorting_normalize_fitness_values[2:population_size]
                                                   - sorting_normalize_fitness_values[0:population_size - 2])
        re_sorting = np.argsort(sorting_normalized_values_index)  # re_sorting to the original order
        matrix_for_crowding[:, i] = crowding_results[re_sorting]

    crowding_distance = np.sum(matrix_for_crowding, axis=1)  # crowding distance of each solution
    return crowding_distance


def remove_using_crowding(fitness_values: np.ndarray, number_solutions_needed: int) -> np.ndarray:
    rn = np.random  # addition
    pop_index = np.arange(fitness_values.shape[0])
    crowding_distance = crowding_calculation(fitness_values)
    selected_pop_index = np.zeros(number_solutions_needed)
    selected_fitness_values = np.zeros((number_solutions_needed, len(fitness_values[0, :])))

    for i in range(number_solutions_needed):
        pop_size = pop_index.shape[0]
        solution_1 = rn.randint(0, pop_size - 1)
        solution_2 = rn.randint(0, pop_size - 1)
        if crowding_distance[solution_1] >= crowding_distance[solution_2]:
            # solution 1 is better than solution 2
            selected_pop_index[i] = pop_index[solution_1]
            selected_fitness_values[i, :] = fitness_values[solution_1, :]
            pop_index = np.delete(pop_index, solution_1, axis=0)  # remove the selected solution
            fitness_values = np.delete(fitness_values, solution_1,
                                       axis=0)  # remove the fitness of the selected solution
            crowding_distance = np.delete(crowding_distance, solution_1,
                                          axis=0)  # remove the related crowding distance

        else:
            # solution 2 is better than solution 1
            selected_pop_index[i] = pop_index[solution_2]
            selected_fitness_values[i, :] = fitness_values[solution_2, :]
            pop_index = np.delete(pop_index, solution_2, axis=0)
            fitness_values = np.delete(fitness_values, solution_2, axis=0)
            crowding_distance = np.delete(crowding_distance, solution_2, axis=0)

    selected_pop_index = np.asarray(selected_pop_index, dtype=int)  # Convert the data to integer
    return selected_pop_index


def selection(all_fitness_values: np.ndarray, selected_size: int) -> np.ndarray:
    remaining_population_idx = np.arange(len(all_fitness_values))
    pareto_indices = np.empty((0,), dtype=int)
    while len(pareto_indices) < selected_size:
        new_pareto_front_idx = find_pareto_front_points(costs=all_fitness_values[remaining_population_idx],
                                                        return_index=True)
        total_pareto_size = len(pareto_indices) + len(new_pareto_front_idx)

        # check the size of pareto front, if larger than self.population_size,
        # remove some solutions using crowding criterion
        if total_pareto_size > selected_size:
            number_solutions_needed = selected_size - len(pareto_indices)
            selected_solutions = remove_using_crowding(all_fitness_values[new_pareto_front_idx],
                                                       number_solutions_needed)
            new_pareto_front_idx = new_pareto_front_idx[selected_solutions]
        pareto_indices = np.hstack((pareto_indices, remaining_population_idx[new_pareto_front_idx]))
        remaining_population_idx = np.setdiff1d(remaining_population_idx, pareto_indices)
    return pareto_indices


def array_divide(topo, lx, ly, lz, divide_number, ini_pop, end_pop):
    lxx = lx / divide_number
    lyy = ly / divide_number
    lzz = lz / divide_number
    topo = topo.reshape((end_pop, lzz, lyy, lxx))
    topo_divided = np.zeros((end_pop, lz, ly, lx))
    for q in range(ini_pop, end_pop + 1):
        for i in range(lx):
            for j in range(ly):
                for k in range(lz):
                    topo_divided[q - 1][k][j][i] = topo[q - 1][k // divide_number][j // divide_number][
                        i // divide_number]
    return topo_divided


def filter_process(topo_divided, sigma, threshold, lx, ly, lz, ini_pop, end_pop):
    topo_filtered = np.zeros((end_pop, lz, ly, lx))
    for q in range(ini_pop, end_pop):
        topo_divided2 = gaussian_filter(topo_divided[q - 1], sigma=sigma)
        for i in range(lx):
            for j in range(ly):
                for k in range(lz):
                    if topo_divided2[k][j][i] >= threshold:
                        topo_divided2[k][j][i] = 1
                    else:
                        topo_divided2[k][j][i] = 0
        topo_filtered[q - 1] = topo_divided2
    return topo_filtered


def quaver_to_full(quaver):
    quarter = np.concatenate((np.flip(quaver, axis=0), quaver), axis=0)
    half = np.concatenate((np.flip(quarter, axis=1), quarter), axis=1)
    full = np.concatenate((np.flip(half, axis=2), half), axis=2)
    return np.swapaxes(full, axis1=0, axis2=2)


def visualize_one_cube(cube_3d_array, full=False):
    fig, ax = plt.subplots(1, 1, figsize=(10, 10), subplot_kw={
        'projection': '3d'
    })
    ax.set(ylabel='y', xticks=[], zticks=[])
    ax.voxels(quaver_to_full(cube_3d_array) if full else cube_3d_array, facecolor='#e02050', edgecolors='k')
    ax.grid(True)
    plt.show()


def visualize_n_cubes(arr_4d, full=False):
    n_row = 1
    n_col = arr_4d.shape[0]
    fig, ax = plt.subplots(n_row, n_col, figsize=(20, 5), subplot_kw={
        'projection': '3d'
    })
    for idx in range(n_col):
        ax[idx].set(xlabel='x', ylabel='y', zlabel='z')
        ax[idx].voxels(quaver_to_full(arr_4d[idx]) if full else arr_4d[idx],
                       facecolor='r', edgecolors='k')
        ax[idx].grid(True)
    plt.show()



from . import Step1_Randomly_generates_parent_topologies as step1
from . import Step2_Generate_offspring_topologies as step2
from . import Step3_Evaluate_fitness_values_and_select as step3
from . import full_steps



from scipy.ndimage import label
from numba import njit, int32
import numpy as np
from typing import Union, Tuple


def make_3d_print_without_support(arr_3d: np.ndarray, max_distance: int = 1) -> int:
    x_size, y_size, z_size = arr_3d.shape
    total_changed_voxels = 0
    while True:
        changed_voxels = 0
        for y_direction in (1, -1):
            for y_idx in range(1, y_size) if y_direction == 1 else range(y_size - 2, -1, -1):
                # Declaration of arr_2d_quarter and labeled_arr
                labeled_arr, max_island_idx = label(arr_3d[:, y_idx, :])
                dead_islands, survived_islands = dead_and_survived_islands(y_idx=y_idx, y_direction=y_direction,
                                                                           x_size=x_size, z_size=z_size,
                                                                           max_island_idx=max_island_idx,
                                                                           labeled_arr=labeled_arr, arr_3d=arr_3d)
                # Eliminating bad voxels by dead islands and survived islands with bad angle
                changed_voxels = voxel_elimination_by_islands(
                    x_size=x_size, z_size=z_size, labeled_arr=labeled_arr, dead_islands=np.array(list(dead_islands)),
                    survived_islands=np.array(list(survived_islands)), arr_3d=arr_3d, y_idx=y_idx,
                    max_distance=max_distance, y_direction=y_direction)
                total_changed_voxels += changed_voxels
        if changed_voxels == 0:
            break
    return total_changed_voxels


@njit
def dead_and_survived_islands(y_idx: int, y_direction: int, x_size: int, z_size: int, max_island_idx: int,
                              labeled_arr: np.ndarray, arr_3d: np.ndarray) -> Tuple[set, set]:
    survived_islands = set()
    dead_islands = set()
    # Determining which ones are the dead islands
    for x_idx in range(x_size):
        for z_idx in range(z_size):
            island_label_num = labeled_arr[x_idx, z_idx]
            if island_label_num and arr_3d[x_idx, y_idx - y_direction, z_idx]:
                survived_islands.add(island_label_num)
    for island_label_num in range(1, max_island_idx + 1):
        if island_label_num not in survived_islands:
            dead_islands.add(island_label_num)
    return dead_islands, survived_islands


@njit
def voxel_elimination_by_islands(x_size: int, z_size: int, labeled_arr: np.ndarray, dead_islands: np.ndarray,
                                 survived_islands: np.ndarray, arr_3d: np.ndarray, y_idx: int,
                                 max_distance: int, y_direction: int) -> int:
    changed_voxels = 0
    for x_idx in range(x_size):
        for z_idx in range(z_size):
            island_idx = labeled_arr[x_idx, z_idx]
            if island_idx in dead_islands:
                arr_3d[x_idx, y_idx, z_idx] = 0
                changed_voxels -= 1
            if island_idx in survived_islands:
                is_any_around = False
                for x_distance in range(-max_distance, max_distance + 1):
                    for z_distance in range(-max_distance, max_distance + 1):
                        x_position = x_idx + x_distance
                        z_position = z_idx + z_distance
                        if x_position < 0:
                            x_position = 0
                        elif x_position > x_size - 1:
                            x_position = x_size - 1
                        if z_position < 0:
                            z_position = 0
                        elif z_position > z_size - 1:
                            z_position = z_size - 1
                        if arr_3d[x_position, y_idx - y_direction, z_position]:
                            is_any_around = True
                if not is_any_around:
                    # print(f'Overhang island Eliminated (x,y,z) = ({x_idx},{y_idx},{z_idx})')
                    arr_3d[x_idx, y_idx, z_idx] = 0
                    changed_voxels -= 1
    return changed_voxels


def one_connected_tree(arr_3d: np.ndarray) -> Union[int, None]:
    arr_shape = arr_3d.shape
    labeled_arr, max_label_idx = label(arr_3d)
    survived_labels = survived_tree_labels(labeled_arr, arr_shape[0], arr_shape[1], arr_shape[2])
    if len(survived_labels) == 0:
        return None
    else:
        unique_arr = np.array(np.unique(labeled_arr, return_counts=True))
        unique_sub_arr = unique_arr[:, np.isin(unique_arr[0], list(map(int, survived_labels)))]
        survived_label = unique_sub_arr[0, np.argmax(unique_sub_arr[1])]
        changed_voxels = one_survived_tree(arr_3d=arr_3d, labeled_arr=labeled_arr, survived_label=survived_label)
        return changed_voxels


@njit
def survived_tree_labels(labeled_arr: np.ndarray, lx: int, ly: int, lz: int) -> set:
    labels_plane_1 = set()
    labels_plane_2 = set()
    labels_plane_3 = set()
    labels_plane_4 = set()
    labels_plane_5 = set()
    labels_plane_6 = set()

    for idx_1 in range(ly):
        for idx_2 in range(lz):
            labels_plane_1.add(labeled_arr[0, idx_1, idx_2])
    for idx_1 in range(ly):
        for idx_2 in range(lz):
            labels_plane_2.add(labeled_arr[lx - 1, idx_1, idx_2])
    for idx_1 in range(lx):
        for idx_2 in range(lz):
            labels_plane_3.add(labeled_arr[idx_1, 0, idx_2])
    for idx_1 in range(lx):
        for idx_2 in range(lz):
            labels_plane_4.add(labeled_arr[idx_1, ly - 1, idx_2])
    for idx_1 in range(lx):
        for idx_2 in range(ly):
            labels_plane_5.add(labeled_arr[idx_1, idx_2, 0])
    for idx_1 in range(lx):
        for idx_2 in range(ly):
            labels_plane_6.add(labeled_arr[idx_1, idx_2, lz - 1])
    return labels_plane_1.intersection(labels_plane_2).intersection(labels_plane_3).intersection(
        labels_plane_4).intersection(labels_plane_5).intersection(labels_plane_6).difference({int32(0)})


@njit
def one_survived_tree(arr_3d: np.ndarray, labeled_arr: np.ndarray, survived_label: int) -> int:
    lx = arr_3d.shape[0]
    ly = arr_3d.shape[1]
    lz = arr_3d.shape[2]
    changed_voxels = 0
    for idx_1 in range(lx):
        for idx_2 in range(ly):
            for idx_3 in range(lz):
                if labeled_arr[idx_1, idx_2, idx_3] == 0:
                    continue
                elif labeled_arr[idx_1, idx_2, idx_3] != survived_label:
                    changed_voxels -= 1
                    arr_3d[idx_1, idx_2, idx_3] = 0
    return changed_voxels


def mutate_and_validate_topology(arr_3d: np.ndarray, mutation_probability: float) -> Union[None, np.ndarray]:
    arr_3d_mutated, voxels_mutation = mutation(arr_3d.copy(), mutation_probability=mutation_probability)
    lx, ly, lz = arr_3d.shape
    while True:
        voxels_validation = 0
        while True:
            _arr_3d_mutated = arr_3d_mutated.copy()
            voxels_oct = one_connected_tree(arr_3d_mutated)
            if voxels_oct is None:
                return None
            voxels_vsc = make_voxels_surface_contact(arr_3d_mutated)
            voxels_3pws = make_3d_print_without_support(arr_3d_mutated)
            voxels_validation += voxels_oct + voxels_3pws + voxels_vsc
            if np.array_equal(arr_3d_mutated, _arr_3d_mutated):
                if voxels_oct == 0 and voxels_3pws == 0 and voxels_vsc == 0:
                    total_voxels_change = voxels_mutation + voxels_validation
                    print('> Validation results')
                    print("- Volume fraction [{}]%".format(100 * np.sum(arr_3d_mutated) / (lx * ly * lz)), end=' | ')
                    print("Changed volume fraction: [{:.2f}]%".format(100 * total_voxels_change / (lx * ly * lz)))
                    return arr_3d_mutated
                else:
                    arr_3d_mutated, voxels_mutation = mutation(arr_3d, mutation_probability=mutation_probability)
                    break


@njit
def mutation(arr_3d: np.ndarray, mutation_probability: float) -> Tuple[np.ndarray, int]:
    _arr_3d = arr_3d.copy()
    lx, ly, lz = arr_3d.shape
    changed_voxels = 0
    for idx_1 in range(lx):
        for idx_2 in range(ly):
            for idx_3 in range(lz):
                if np.random.random() < mutation_probability:
                    if _arr_3d[idx_1, idx_2, idx_3] == 1:
                        _arr_3d[idx_1, idx_2, idx_3] = 0
                        changed_voxels -= 1
                    else:
                        _arr_3d[idx_1, idx_2, idx_3] = 1
                        changed_voxels += 1
    return _arr_3d, changed_voxels


@njit
def make_voxels_surface_contact(arr_3d: np.ndarray) -> int:
    nx, ny, nz = arr_3d.shape
    total_changed_voxels = 0
    point_to_point = np.array([[1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1],
                               [-1, -1, 1], [-1, 1, -1], [1, -1, -1], [-1, -1, -1]])
    edge_to_edge = np.array([[0, 1, 1], [0, -1, 1], [0, 1, -1], [0, -1, -1], [1, 0, 1], [-1, 0, 1],
                             [1, 0, -1], [-1, 0, -1], [1, 1, 0], [-1, 1, 0], [1, -1, 0], [-1, -1, 0]])
    while True:
        changed_voxels = 0
        for x in range(nx):
            for y in range(ny):
                for z in range(nz):
                    if arr_3d[x, y, z] == 0:
                        continue
                    for xx, yy, zz in point_to_point:
                        px, py, pz = x + xx, y + yy, z + zz
                        if px < 0 or py < 0 or pz < 0 or px == nx or py == ny or pz == nz:
                            continue
                        if arr_3d[px, py, pz] == 0:
                            continue
                        if arr_3d[px, y, z] or arr_3d[x, py, z] or arr_3d[x, y, pz]:  # 100 010 001
                            continue
                        if arr_3d[px, py, z] or arr_3d[px, y, pz] or arr_3d[x, py, pz]:  # 110 101 011
                            continue
                        routine = np.random.randint(6)
                        changed_voxels += 2
                        if routine == 0:
                            arr_3d[px, y, z] = 1  # 100
                            arr_3d[px, py, z] = 1  # 110
                        elif routine == 1:
                            arr_3d[px, y, z] = 1  # 100
                            arr_3d[px, y, pz] = 1  # 101
                        elif routine == 2:
                            arr_3d[x, py, z] = 1  # 010
                            arr_3d[px, py, z] = 1  # 110
                        elif routine == 3:
                            arr_3d[x, py, z] = 1  # 010
                            arr_3d[x, py, pz] = 1  # 011
                        elif routine == 4:
                            arr_3d[x, y, pz] = 1  # 001
                            arr_3d[px, y, pz] = 1  # 101
                        else:
                            arr_3d[x, y, pz] = 1  # 001
                            arr_3d[x, py, pz] = 1  # 011
                    for xx, yy, zz in edge_to_edge:
                        px, py, pz = x + xx, y + yy, z + zz
                        if px < 0 or py < 0 or pz < 0 or px == nx or py == ny or pz == nz:
                            continue
                        if arr_3d[px, py, pz] == 0:
                            continue
                        if xx == 0:  # 011
                            if arr_3d[x, y, pz] or arr_3d[x, py, z]:
                                continue
                            routine = np.random.randint(2)
                            changed_voxels += 1
                            if routine == 0:
                                arr_3d[x, y, pz] = 1  # 001
                            else:
                                arr_3d[x, py, z] = 1  # 010
                        elif yy == 0:  # 101
                            if arr_3d[px, y, z] or arr_3d[x, y, pz]:
                                continue
                            routine = np.random.randint(2)
                            changed_voxels += 1
                            if routine == 0:
                                arr_3d[px, y, z] = 1  # 100
                            else:
                                arr_3d[x, y, pz] = 1  # 001
                        else:  # 110
                            if arr_3d[px, y, z] or arr_3d[x, py, z]:
                                continue
                            routine = np.random.randint(2)
                            changed_voxels += 1
                            if routine == 0:
                                arr_3d[px, y, z] = 1  # 100
                            else:
                                arr_3d[x, py, z] = 1  # 010
        total_changed_voxels += changed_voxels
        if changed_voxels == 0:
            break
    return total_changed_voxels



"""
Randomly generate parent topologies
"""


def run():
    from ..GeneticAlgorithm import random_parent_generation
    from ..ParameterDefinitions import Parameters
    from ..PostProcessing import visualize_one_cube
    import os
    script_dir = os.path.dirname(__file__)
    sample_data_dir = os.path.join(script_dir, 'sample_data')
    os.chdir(sample_data_dir)

    # Parameters
    number_of_voxels_x = 5
    number_of_voxels_y = 5
    number_of_voxels_z = 5
    topology_density = 0.3
    number_of_topologies_to_create = 10
    mutation_rate = 0.1
    save_parent_topologies_file_as = 'Topologies_1'
    show_created_parent_topologies = True
    full_sized_cube = True

    # Initialize parameters
    parameters = Parameters(lx=number_of_voxels_x, ly=number_of_voxels_y, lz=number_of_voxels_z,
                            end_pop=number_of_topologies_to_create, mutation_rate=mutation_rate)
    parameters.post_initialize()

    # Create parent topologies and visualize as full-sized cube
    parent_topologies = random_parent_generation(density=topology_density, params=parameters,
                                                 save_file_as=save_parent_topologies_file_as)
    if show_created_parent_topologies:
        for parent_topology in parent_topologies:
            visualize_one_cube(parent_topology, full=full_sized_cube)


if __name__ == '__main__':
    run()



from setuptools import setup, find_packages

long_descriptions = open('README.md', encoding='utf-8').read().encode("ascii", "ignore").decode('ascii')
print(long_descriptions)

setup(name="auxeticmop",
      version='1.0.1',
      description="A package for finding meta-material structure using ABAQUS and MOP evolutionary algorithm approaches.",
      author='Seongbin Choi',
      author_email='dcas@naver.com',
      url='https://github.com/c0sogi/AuxeticMOP-with-ABAQUS',
      long_description=long_descriptions,
      long_description_content_type='text/markdown',
      license='MIT',
      python_requires='>=3.7,<3.11',
      install_requires=['numpy', 'numba', 'aiofiles', 'matplotlib', 'scipy'],
      packages=find_packages(),
      package_data={"": ["sample_scripts/sample_data/*"]},
      zip_safe=False,
      include_package_data=True,
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: Microsoft :: Windows",
      ],
      )



from dataclasses import dataclass
import numpy as np
from typing import Union


@dataclass  # Use @dataclass(kw_only=True) for Python version >= 3.10
class Parameters:
    abaqus_mode: str = 'script'  # noGUI: without abaqus gui, script: with abaqus gui
    # mode: str = 'GA'  # GA mode
    evaluation_version: str = 'ver5'  # fitness value evaluation mode
    end_pop: int = 10  # Last population number
    end_gen: int = 50  # Last generation number
    mutation_rate: float = 0.1  # mutation process option
    unit_l: float = 3  # Voxel length
    lx: int = 5  # Number of voxels in x-direction
    ly: int = 5  # Number of voxels in y-direction
    lz: int = 5  # Number of voxels in z-direction
    divide_number: int = 1  # up-scaling factor
    mesh_size: float = 1.0  # abaqus meshing option
    dis_y: float = -0.005  # abaqus boundary condition option
    material_modulus: float = 1100  # abaqus material property option
    # poisson_ratio: float = 0.4  # abaqus material property option
    # density: float = 1  # abaqus material property option
    penalty_coefficient: float = 0.1  # fitness value evaluation option
    # sigma: float = 1  # filtering option
    # threshold: float = 0.5  # filtering option
    n_cpus: int = 1  # abaqus option
    n_gpus: int = 0  # abaqus option

    def post_initialize(self):  # call this method to set initial values to real value to be used
        self.lx *= self.divide_number
        self.ly *= self.divide_number
        self.lz *= self.divide_number  # number of voxels after increasing resolution
        self.unit_l /= self.divide_number
        self.mesh_size *= self.unit_l
        self.dis_y *= self.ly * self.unit_l  # boundary condition (displacement)


@dataclass
class JsonFormat:
    start_topology_from: Union[int, None]
    topologies_file_name: str
    topologies_key: str
    exit_abaqus: bool


@dataclass
class GuiParameters:
    parameter_file_name: str = '_PARAMETERS_'
    padx: int = 5  # Padding width
    pady: int = 5  # Padding height
    left_width: int = 1400  # default width: 400
    right_width: int = 400
    height: int = 750
    button_width: int = 15
    polling_rate: float = 10.0
    title: str = 'Abaqus-Python Unified Control Interface'


radiobutton_name_dict = {
    'abaqus_mode': ('noGUI', 'script'),
    'mode': ('GA', 'random'),
    'evaluation_version': ('ver3', 'ver4', 'ver5')
}


@dataclass
class FitnessDefinitions:
    vars_definitions: dict
    fitness_value_definitions: Union[tuple, list]


material_property_definitions = {
    'material_name': 'resin',
    'density': 1.2e-09,
    'engineering_constants': (1500, 1200, 1500, 0.35, 0.35, 0.35, 450, 550, 450)
}

exported_field_outputs_format = {
    'displacement': {'xMax': np.ndarray, 'yMax': np.ndarray, 'zMax': np.ndarray},
    'rotation': np.ndarray,
    'reaction_force': np.ndarray,
    'mises_stress': {'max': float, 'min': float, 'average': float}
}

fitness_definitions = {
    'ver1': FitnessDefinitions(
        vars_definitions={
            'dis11': ('displacement', 'xMax', 0),
            'dis22': ('displacement', 'yMax', 1),
            'rf22': ('reaction_force', 2),
            'max_rf22': '$max_rf22',
            'k': '@penalty_coefficient',  # The prefix @ means this is variable is from Parameters
            'total_voxels': '$total_voxels',  # The prefix $ means this is predefined variable
            'lx': '@lx',
            'ly': '@ly',
            'lz': '@lz'
        },
        fitness_value_definitions=(
            '(rf22 / max_rf22) + k * total_voxels / (lx * ly * lz)',
            '- (dis11 / dis22) + k * total_voxels / (lx * ly * lz)'
        )),
    'ver2': FitnessDefinitions(
        vars_definitions={
            'rf22': ('reaction_force', 2),
            'max_rf22': '$max_rf22',  # The prefix @ means this is variable is from Parameters
            'total_voxels': '$total_voxels',  # The prefix $ means this is predefined variable
            'lx': '@lx',
            'ly': '@ly',
            'lz': '@lz'
        },
        fitness_value_definitions=(
            'total_voxels / (lx * ly * lz)',
            'rf22 / max_rf22'
        )),
    'ver3': FitnessDefinitions(
        vars_definitions={
            'dis11': ('displacement', 'xMax', 0),
            'dis22': ('displacement', 'yMax', 1),
            'dis33': ('displacement', 'zMax', 2),
            'k': '@penalty_coefficient',  # The prefix @ means this is variable is from Parameters
            'total_voxels': '$total_voxels',  # The prefix $ means this is predefined variable
            'lx': '@lx',
            'ly': '@ly',
            'lz': '@lz'
        },
        fitness_value_definitions=(
            '- (dis11 / dis22) + k * (total_voxels) / (lx * ly * lz)',
            '- (dis33 / dis22) + k * (total_voxels) / (lx * ly * lz)'
        )),
    'ver4': FitnessDefinitions(
        vars_definitions={
            'max_mises': ('mises_stress', 'max'),
            'total_voxels': '$total_voxels',  # The prefix $ means this is predefined variable
            'lx': '@lx',  # The prefix @ means this is variable is from Parameters
            'ly': '@ly',
            'lz': '@lz'
        },
        fitness_value_definitions=(
            'max_mises',
            'total_voxels / (lx * ly * lz)'
        )),
    'ver5': FitnessDefinitions(
        vars_definitions={
            'max_mises': ('mises_stress', 'max'),
            'dis11': ('displacement', 'xMax', 0),
            'dis22': ('displacement', 'yMax', 1),
            'dis33': ('displacement', 'zMax', 2)
        },
        fitness_value_definitions=(
            'max_mises',
            'max(-dis11/dis22, -dis33/dis22)'
        ))
}

translate_dictionary = {'abaqus_script': 'Filename of ABAQUS script',
                        'abaqus_mode': 'ABAQUS execution mode',
                        'mode': 'GA Mode',
                        'evaluation_version': 'GA evaluation version',
                        'restart_pop': '[P] Restart from population',
                        'ini_pop': '[P] First Population',
                        'end_pop': '[P] Populations per generation',
                        'ini_gen': '[G] First Generation',
                        'end_gen': '[G] Last Generation',
                        'mutation_rate': 'Mutation rate(0~1)',
                        'unit_l': 'Voxel unit length(mm)',
                        'lx': 'Voxel number in X-direction',
                        'ly': 'Voxel number in Y-direction',
                        'lz': 'Voxel number in Z-direction',
                        'divide_number': 'Upscale multiplier(1~)',
                        'mesh_size': 'Mesh size/Voxel size(0~1)',
                        'dis_y': 'Y Compression ratio(-1~1)',
                        'material_modulus': "Young's modulus(MPa)",
                        'poisson_ratio': "Poisson's ratio(0~1)",
                        'density': 'Material density(ton/mm3)',
                        'MaxRF22': 'Maximum RF22(N)',
                        'penalty_coefficient': 'Penalty coefficient',
                        'sigma': 'Sigma for filtering',
                        'threshold': 'Threshold for filtering',
                        'n_cpus': 'CPU cores for abaqus',
                        'n_gpus': 'GPU cores for abaqus',
                        'timeout': 'Timeout of validation process(s)'}
